package org.openmuc.android.activities;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.openmuc.android.R;
import org.openmuc.android.rest.RestClient;
import org.openmuc.android.rest.data.Channel;
import org.openmuc.android.util.ChannelAdapter;
import org.openmuc.android.util.SharedPreferenceUtil;

import android.annotation.TargetApi;
import android.app.Activity;
import android.content.Intent;
import android.content.SharedPreferences;
import android.os.AsyncTask;
import android.os.Build;
import android.os.Bundle;
import android.preference.PreferenceManager;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.widget.CheckBox;
import android.widget.LinearLayout;
import android.widget.ListView;
import android.widget.TextView;

/**
 * Activity used to display and add channels that are available on the server and not already included in the list of
 * channels that are read by the app
 * 
 * @author Karsten Müller-Bier
 */
public class AddChannelActivity extends Activity {

	public static final int REQUEST_CODE = 1;

	RestClient client;

	LoadChannelTask runningTask;

	List<Channel> alreadyRegisteredChannel;
	List<Channel> channelList;
	ChannelAdapter adapter;

	@Override
	protected void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		setContentView(R.layout.activity_add_channel);

		setupActionBar();
	}

	@Override
	protected void onStart() {
		super.onStart();

		SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(this);
		String serverAddress = prefs.getString(getString(R.string.preference_server_address), "");
		String serverPort = prefs.getString(getString(R.string.preference_server_port), "");
		client = new RestClient(serverAddress + ":" + serverPort);

		Set<String> registeredChannels = SharedPreferenceUtil.getChannelLabels(this);
		alreadyRegisteredChannel = new ArrayList<Channel>(registeredChannels.size());
		for (String label : registeredChannels) {
			Channel c = new Channel();
			c.setLabel(label);
			alreadyRegisteredChannel.add(c);
		}

		channelList = Collections.emptyList();
		adapter = new ChannelAdapter(this, R.layout.selectable_row, channelList);
		ListView view = (ListView) findViewById(R.id.addChannelList);
		view.setAdapter(adapter);

		runningTask = new LoadChannelTask();
		runningTask.execute(client);
	}

	@Override
	protected void onStop() {
		super.onStop();

		client.release();
		if (runningTask != null) {
			runningTask.cancel(true);
		}
	}

	/**
	 * Set up the {@link android.app.ActionBar}, if the API is available.
	 */
	@TargetApi(Build.VERSION_CODES.HONEYCOMB)
	private void setupActionBar() {
		if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.HONEYCOMB) {
			getActionBar().setDisplayHomeAsUpEnabled(true);
		}
	}

	@Override
	public boolean onCreateOptionsMenu(Menu menu) {
		// Inflate the menu; this adds items to the action bar if it is present.
		getMenuInflater().inflate(R.menu.add_channel, menu);
		return true;
	}

	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
		Intent returnIntent = new Intent();
		switch (item.getItemId()) {
		case android.R.id.home:
			// This ID represents the Home or Up button. In the case of this
			// activity, the Up button is shown. Use NavUtils to allow users
			// to navigate up one level in the application structure. For
			// more details, see the Navigation pattern on Android Design:
			//
			// http://developer.android.com/design/patterns/navigation.html#up-vs-back
			//
			setResult(RESULT_CANCELED, returnIntent);
			finish();
			return true;
		case R.id.action_accept:
			Set<String> addedChannels = getCheckedChannels();
			returnIntent.putExtra("channelLabels", addedChannels.toArray(new String[0]));
			setResult(RESULT_OK, returnIntent);
			finish();
			return true;
		case R.id.action_refresh:
			runningTask = new LoadChannelTask();
			runningTask.execute(client);
			return true;
		}
		return super.onOptionsItemSelected(item);
	}

	/**
	 * Traverses the channelList returns the label of all channels that were checked
	 * 
	 * @return Set of labels of all checked channels
	 */
	private Set<String> getCheckedChannels() {
		Set<String> result = new HashSet<String>();

		ListView view = (ListView) findViewById(R.id.addChannelList);
		int listSize = view.getChildCount();
		for (int i = 0; i < listSize; i++) {
			LinearLayout listItem = (LinearLayout) view.getChildAt(i);
			CheckBox check = (CheckBox) listItem.findViewById(R.id.channelselected);
			if (check.isChecked()) {
				TextView text = (TextView) listItem.findViewById(R.id.toptext);
				result.add(text.getText().toString());
			}
		}

		return result;
	}

	/**
	 * AsyncTask that scans all available channels of the server and displays all channels not already registered
	 * 
	 * @author Karsten Müller-Bier
	 */
	private class LoadChannelTask extends AsyncTask<RestClient, Void, List<Channel>> {

		@Override
		protected List<Channel> doInBackground(RestClient... params) {
			try {
				return params[0].listChannels();
			} catch (IOException e) {
				return Collections.emptyList();
			}
		}

		@Override
		protected void onPostExecute(List<Channel> result) {
			result.removeAll(alreadyRegisteredChannel);
			channelList = result;
			adapter = new ChannelAdapter(AddChannelActivity.this, R.layout.selectable_row, channelList);
			ListView view = (ListView) findViewById(R.id.addChannelList);
			view.setAdapter(adapter);

			findViewById(R.id.LoadingBar).setVisibility(View.GONE);
			findViewById(R.id.LoadingText).setVisibility(View.GONE);
			if (channelList.size() == 0) {
				findViewById(R.id.errorText).setVisibility(View.VISIBLE);
				findViewById(R.id.addChannelList).setVisibility(View.GONE);
			}
			else {
				findViewById(R.id.errorText).setVisibility(View.GONE);
				findViewById(R.id.addChannelList).setVisibility(View.VISIBLE);
			}

			runningTask = null;
		}
	}
}
