package org.openmuc.android.activities.handheld;

import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import org.openmuc.android.R;
import org.openmuc.android.fragments.ChannelChartFragment;
import org.openmuc.android.fragments.TimespanSelectorFragment;
import org.openmuc.android.fragments.TimespanSelectorFragment.ChartCallback;
import org.openmuc.android.rest.RestClient;
import org.openmuc.android.rest.data.Channel;
import org.openmuc.android.rest.data.Record;
import org.openmuc.android.util.ChannelRecordSeries;
import org.openmuc.android.util.ChannelRecordUpdater;
import org.openmuc.android.util.ChannelRecordUpdater.RecordUpdaterCallback;

import android.annotation.TargetApi;
import android.content.Intent;
import android.content.SharedPreferences;
import android.os.AsyncTask;
import android.os.Build;
import android.os.Bundle;
import android.preference.PreferenceManager;
import android.support.v4.app.FragmentActivity;
import android.support.v4.app.FragmentManager;
import android.support.v4.app.FragmentTransaction;
import android.support.v4.app.NavUtils;
import android.view.Menu;
import android.view.MenuItem;
import android.widget.RelativeLayout;

/**
 * Smartphone activity showing a historical graph of a single selected channel sent with the intent
 * 
 * @author Karsten Müller-Bier
 */
public class ChannelChart extends FragmentActivity implements ChartCallback, RecordUpdaterCallback {

	private Channel channel;

	private ChannelChartFragment plotFragment;
	private TimespanSelectorFragment timespanFragment;

	private ChannelRecordUpdater runner;
	private ChannelRecordSeries series;

	private RestClient client;

	@Override
	protected void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		setContentView(R.layout.activity_channel_chart);
		// Show the Up button in the action bar.
		setupActionBar();

		List<Record> records;
		if (savedInstanceState == null) {
			plotFragment = new ChannelChartFragment();
			timespanFragment = new TimespanSelectorFragment();

			FragmentManager fm = getSupportFragmentManager();
			FragmentTransaction tr = fm.beginTransaction();

			tr.add(R.id.timespanSelectorFragment, timespanFragment);
			tr.add(R.id.channelChartFragment, plotFragment);
			tr.commit();

			records = Collections.emptyList();
		}
		else {
			FragmentManager fm = getSupportFragmentManager();
			timespanFragment = (TimespanSelectorFragment) fm.findFragmentById(R.id.timespanSelectorFragment);
			plotFragment = (ChannelChartFragment) fm.findFragmentById(R.id.channelChartFragment);

			records = Arrays.asList((Record[]) savedInstanceState.getParcelableArray("records"));
		}
		series = new ChannelRecordSeries("", null, records);

		Intent receivedIntent = getIntent();
		String channelLabel = receivedIntent.getStringExtra("channel");
		channel = new Channel();
		channel.setLabel(channelLabel);

		setTitle(getTitle() + " " + channelLabel);
	}

	@Override
	protected void onStart() {
		super.onStart();

		SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(this);
		String serverAddress = prefs.getString(getString(R.string.preference_server_address), "");
		String serverPort = prefs.getString(getString(R.string.preference_server_port), "");
		client = new RestClient(serverAddress + ":" + serverPort);
		runner = new ChannelRecordUpdater(client);
	}

	@Override
	protected void onPause() {
		super.onPause();
	}

	@Override
	protected void onSaveInstanceState(Bundle outState) {
		super.onSaveInstanceState(outState);

		List<Record> records = series.slice(series.getEarliestTimestamp() - 1, series.getLatestTimestamp() + 1);
		outState.putParcelableArray("records", records.toArray(new Record[0]));
	}

	@Override
	protected void onStop() {
		super.onStop();
		runner.stop();
		client.release();
	}

	/**
	 * Set up the {@link android.app.ActionBar}, if the API is available.
	 */
	@TargetApi(Build.VERSION_CODES.HONEYCOMB)
	private void setupActionBar() {
		if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.HONEYCOMB) {
			getActionBar().setDisplayHomeAsUpEnabled(true);
		}
	}

	@Override
	public boolean onCreateOptionsMenu(Menu menu) {
		// Inflate the menu; this adds items to the action bar if it is present.
		getMenuInflater().inflate(R.menu.channel_chart, menu);
		return true;
	}

	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
		switch (item.getItemId()) {
		case android.R.id.home:
			// This ID represents the Home or Up button. In the case of this
			// activity, the Up button is shown. Use NavUtils to allow users
			// to navigate up one level in the application structure. For
			// more details, see the Navigation pattern on Android Design:
			//
			// http://developer.android.com/design/patterns/navigation.html#up-vs-back
			//
			runner.stop();
			NavUtils.navigateUpFromSameTask(this);
			return true;
		}
		return super.onOptionsItemSelected(item);
	}

	@Override
	public void timeSpanChanged() {
		runner.stop();
		plotFragment.showPlot(false);

		Long duration = timespanFragment.getDuration();
		if (duration == null) {
			new LoadHistoryTask(client, timespanFragment.getStartTimestamp(), timespanFragment.getEndTimestamp())
					.execute(channel);
		}
		else {
			new LoadHistoryTask(client, duration).execute(channel);
		}
	}

	/**
	 * AsyncTask used to load the history of a channel at the given time frame. If the task was successful, the graph of
	 * the channel will be displayed on the device and, if set by the user, continuously updated.
	 * 
	 * @author Karsten Müller-Bier
	 */
	private class LoadHistoryTask extends AsyncTask<Channel, Void, List<Record>> {

		private final RestClient client;
		private final Long maxDuration;
		private final Long from;
		private final Long until;
		private String title;

		public LoadHistoryTask(RestClient client, Long maxDuration) {
			this.client = client;
			this.maxDuration = maxDuration;
			until = System.currentTimeMillis();
			from = until - maxDuration;
		}

		public LoadHistoryTask(RestClient client, Long from, Long until) {
			this.client = client;
			maxDuration = null;
			this.from = from;
			this.until = until;
		}

		@Override
		protected List<Record> doInBackground(Channel... params) {
			if (params.length == 0) {
				cancel(true);
				return null;
			}

			Channel channel = params[0];

			List<Record> result = new LinkedList<Record>();
			List<Record> middle = series.slice(from, until);
			List<Record> begin = Collections.emptyList();
			List<Record> end = Collections.emptyList();

			try {
				if (from < series.getEarliestTimestamp()) {
					long untilMiddle = Math.min(until, series.getEarliestTimestamp() - 1);
					begin = client.readChannelHistory(channel, from, untilMiddle);
				}
				if (until > series.getLatestTimestamp()) {
					long fromMiddle = Math.max(from, series.getLatestTimestamp() + 1);
					end = client.readChannelHistory(channel, fromMiddle, until);
				}
			} catch (IOException e) {
				cancel(true);
				return null;
			}

			title = channel.getLabel();
			result.addAll(begin);
			result.addAll(middle);
			result.addAll(end);
			return result;
		}

		@Override
		protected void onPostExecute(List<Record> result) {
			series = new ChannelRecordSeries(title, maxDuration, result);

			plotFragment.clearPlotSeries();
			plotFragment.setPlotSeries(series, 0);
			plotFragment.showPlot(true);

			RelativeLayout layout = (RelativeLayout) findViewById(R.id.chartLayout);
			layout.invalidate();

			if (maxDuration != null) {
				SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(ChannelChart.this);
				int interval = Integer.parseInt(prefs.getString(getString(R.string.pref_key_update_interval), "0"));
				runner.start(channel, ChannelChart.this, interval);
			}
		}
	}

	@Override
	public void newRecord(Record record) {
		series.addRecord(record);
		plotFragment.replot();
	}
}
