package org.openmuc.android.activities.handheld;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.openmuc.android.R;
import org.openmuc.android.activities.AddChannelActivity;
import org.openmuc.android.activities.SettingsActivity;
import org.openmuc.android.fragments.ChannelListFragment;
import org.openmuc.android.fragments.ChannelListFragment.ChannelListFragmentCallback;
import org.openmuc.android.rest.RestClient;
import org.openmuc.android.rest.data.Channel;
import org.openmuc.android.util.ChannelUpdater;
import org.openmuc.android.util.SharedPreferenceUtil;

import android.content.Intent;
import android.content.SharedPreferences;
import android.os.Build;
import android.os.Bundle;
import android.preference.PreferenceManager;
import android.support.v4.app.FragmentActivity;
import android.support.v4.app.FragmentManager;
import android.support.v4.app.FragmentTransaction;
import android.support.v4.app.NavUtils;
import android.view.Menu;
import android.view.MenuItem;

/**
 * Main activity for smartphone devices. Lists all registered channels and updates them regularly.
 * 
 * @author Karsten Müller-Bier
 */
public class ChannelList extends FragmentActivity implements ChannelListFragmentCallback {

	private final List<Channel> channelsToRead = new ArrayList<Channel>();
	private RestClient restClient;

	private ChannelUpdater updater;
	private ChannelListFragment fragment;

	private Integer deleteIconId;

	@Override
	protected void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);

		setContentView(R.layout.activity_channel_list);

		SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(this);
		String serverAddress = prefs.getString(getString(R.string.preference_server_address), "");
		if (serverAddress.isEmpty()) {
			Intent configScreenIntent = new Intent(this, SettingsActivity.class);
			startActivity(configScreenIntent);
		}

		Set<String> labels = SharedPreferenceUtil.getChannelLabels(this);
		for (String label : labels) {
			Channel newChannel = new Channel();
			newChannel.setLabel(label);
			channelsToRead.add(newChannel);
		}

		if (savedInstanceState == null) {
			fragment = new ChannelListFragment();

			FragmentManager fm = getSupportFragmentManager();
			FragmentTransaction tr = fm.beginTransaction();

			tr.add(R.id.channelListFragment, fragment);
			tr.commit();
		}
		else {
			FragmentManager fm = getSupportFragmentManager();
			fragment = (ChannelListFragment) fm.findFragmentById(R.id.channelListFragment);
		}
		fragment.setChannelList(channelsToRead);

		updater = new ChannelUpdater(channelsToRead);
	}

	@Override
	protected void onStart() {
		super.onStart();

		SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(this);
		String serverAddress = prefs.getString(getString(R.string.preference_server_address), "");
		String serverPort = prefs.getString(getString(R.string.preference_server_port), "");

		restClient = new RestClient(serverAddress + ":" + serverPort);
	}

	@Override
	protected void onResume() {
		super.onResume();

		SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(this);
		int interval = Integer.parseInt(prefs.getString(getString(R.string.pref_key_update_interval), "0"));
		updater.start(fragment.getAdapterHandler(), restClient, interval);
	}

	@Override
	protected void onPause() {
		super.onPause();

		updater.stop();
	}

	@Override
	protected void onStop() {
		super.onStop();

		restClient.release();
	}

	/**
	 * Add additional menu item if device runs on Android with version 2.3.3 or lower
	 */
	@Override
	public boolean onCreateOptionsMenu(Menu menu) {
		// Inflate the menu; this adds items to the action bar if it is present.
		getMenuInflater().inflate(R.menu.main, menu);
		if (Build.VERSION.SDK_INT <= Build.VERSION_CODES.HONEYCOMB) {
			deleteIconId = menu.add("Delete Channels").setIcon(R.drawable.content_discard).getItemId();
		}
		return true;
	}

	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
		switch (item.getItemId()) {
		case android.R.id.home:
			// This ID represents the Home or Up button. In the case of this
			// activity, the Up button is shown. Use NavUtils to allow users
			// to navigate up one level in the application structure. For
			// more details, see the Navigation pattern on Android Design:
			//
			// http://developer.android.com/design/patterns/navigation.html#up-vs-back
			//
			NavUtils.navigateUpFromSameTask(this);
			return true;
		case R.id.action_settings:
			Intent goToSettings = new Intent(this, SettingsActivity.class);
			startActivity(goToSettings);
			return true;
		case R.id.action_addchannel:
			Intent goToAddChannels = new Intent(this, AddChannelActivity.class);
			startActivityForResult(goToAddChannels, AddChannelActivity.REQUEST_CODE);
			return true;
		case R.id.action_refresh:
			updater.pollUpdate();
			return true;
		}
		if (deleteIconId != null && item.getItemId() == deleteIconId) {
			Intent goToDeleteChannels = new Intent(this, DeleteChannelActivity.class);
			startActivityForResult(goToDeleteChannels, DeleteChannelActivity.REQUEST_CODE);
			return true;
		}
		return super.onOptionsItemSelected(item);
	}

	@Override
	protected void onActivityResult(int requestCode, int resultCode, Intent data) {
		if (resultCode == RESULT_OK) {
			if (requestCode == AddChannelActivity.REQUEST_CODE) {
				String[] newChannels = data.getStringArrayExtra("channelLabels");
				Set<String> storedChannels = SharedPreferenceUtil.getChannelLabels(this);
				Collections.addAll(storedChannels, newChannels);
				SharedPreferenceUtil.writeChannelLabels(this, storedChannels);

				for (String label : newChannels) {
					Channel newChannel = new Channel();
					newChannel.setLabel(label);
					channelsToRead.add(newChannel);
				}
			}
			else if (requestCode == DeleteChannelActivity.REQUEST_CODE) {
				String[] channelLabels = data.getStringArrayExtra("channelLabels");
				Set<String> channelsToDelete = new HashSet<String>();
				for (String label : channelLabels) {
					channelsToDelete.add(label);
				}
				deleteChannel(channelsToDelete);
			}
		}
	}

	@Override
	public void channelSelected(Channel selectedChannel) {
		Intent gotoChannelChart = new Intent(ChannelList.this, ChannelChart.class);
		gotoChannelChart.putExtra("channel", selectedChannel.getLabel());
		startActivity(gotoChannelChart);
	}

	@Override
	public void deleteChannel(Set<String> channelsToDelete) {
		Set<String> channels = SharedPreferenceUtil.getChannelLabels(this);
		channels.removeAll(channelsToDelete);
		SharedPreferenceUtil.writeChannelLabels(this, channels);

		Iterator<Channel> iter = channelsToRead.iterator();
		while (iter.hasNext()) {
			Channel channel = iter.next();
			if (channelsToDelete.contains(channel.getLabel())) {
				iter.remove();
			}
		}
	}
}
