package org.openmuc.android.activities.tablet;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.openmuc.android.R;
import org.openmuc.android.activities.AddChannelActivity;
import org.openmuc.android.activities.SettingsActivity;
import org.openmuc.android.fragments.ChannelChartFragment;
import org.openmuc.android.fragments.ChannelListFragment;
import org.openmuc.android.fragments.ChannelListFragment.ChannelListFragmentCallback;
import org.openmuc.android.fragments.TimespanSelectorFragment;
import org.openmuc.android.fragments.TimespanSelectorFragment.ChartCallback;
import org.openmuc.android.rest.RestClient;
import org.openmuc.android.rest.data.Channel;
import org.openmuc.android.rest.data.Record;
import org.openmuc.android.util.ChannelRecordSeries;
import org.openmuc.android.util.ChannelUpdater;
import org.openmuc.android.util.ChannelUpdater.UpdaterCallback;
import org.openmuc.android.util.SharedPreferenceUtil;

import android.content.Intent;
import android.content.SharedPreferences;
import android.content.res.Configuration;
import android.os.AsyncTask;
import android.os.Bundle;
import android.preference.PreferenceManager;
import android.support.v4.app.FragmentActivity;
import android.support.v4.app.FragmentManager;
import android.support.v4.app.FragmentTransaction;
import android.support.v4.app.NavUtils;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.widget.RelativeLayout;

/**
 * Main activity for tablet devices. Lists all registered channels, updates them regularly and displays the graph of up
 * to 3 selected channels.
 * 
 * @author Karsten Müller-Bier
 */
public class DashBoard extends FragmentActivity implements ChannelListFragmentCallback, ChartCallback, UpdaterCallback {

	private RestClient restClient;

	private final List<Channel> channelsToRead = Collections.synchronizedList(new ArrayList<Channel>());
	private final List<Channel> selectedChannel = new LinkedList<Channel>();

	private ChannelUpdater updater;

	private Map<Channel, ChannelRecordSeries> channelSeries;

	private ChannelListFragment listFragment;
	private ChannelChartFragment plotFragment;
	private TimespanSelectorFragment timespanFragment;

	private boolean newRecord;

	@Override
	protected void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		setContentView(R.layout.activity_dash_board);

		SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(this);
		String serverAddress = prefs.getString(getString(R.string.preference_server_address), "");
		if (serverAddress.isEmpty()) {
			Intent configScreenIntent = new Intent(this, SettingsActivity.class);
			startActivity(configScreenIntent);
		}

		Set<String> labels = SharedPreferenceUtil.getChannelLabels(this);
		for (String label : labels) {
			Channel newChannel = new Channel();
			newChannel.setLabel(label);
			channelsToRead.add(newChannel);
		}

		showChannelDetails(false);
		updater = new ChannelUpdater(channelsToRead);
		channelSeries = new HashMap<Channel, ChannelRecordSeries>();

		if (savedInstanceState == null) {
			listFragment = new ChannelListFragment();
			plotFragment = new ChannelChartFragment();
			timespanFragment = new TimespanSelectorFragment();

			FragmentManager fm = getSupportFragmentManager();
			FragmentTransaction tr = fm.beginTransaction();

			tr.add(R.id.channelListFragment, listFragment);
			tr.add(R.id.timespanSelectorFragment, timespanFragment);
			tr.add(R.id.channelChartFragment, plotFragment);
			tr.commit();

			channelSeries = new HashMap<Channel, ChannelRecordSeries>();
		}
		else {
			FragmentManager fm = getSupportFragmentManager();
			listFragment = (ChannelListFragment) fm.findFragmentById(R.id.channelListFragment);
			timespanFragment = (TimespanSelectorFragment) fm.findFragmentById(R.id.timespanSelectorFragment);
			plotFragment = (ChannelChartFragment) fm.findFragmentById(R.id.channelChartFragment);

			int numOfSeries = savedInstanceState.getInt("seriesNumber");

			for (int i = 0; i < numOfSeries; i++) {
				String channelLabel = savedInstanceState.getString("seriesLabel_" + i);
				List<Record> records = Arrays.asList((Record[]) savedInstanceState.getParcelableArray("seriesRecords_"
						+ i));
				for (Channel c : channelsToRead) {
					if (c.getLabel().equals(channelLabel)) {
						selectedChannel.add(c);
						channelSeries.put(c, new ChannelRecordSeries(channelLabel, null, records));
					}
				}
			}
		}
		listFragment.setChannelList(channelsToRead);
	}

	@Override
	protected void onStart() {
		super.onStart();

		SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(this);
		String serverAddress = prefs.getString(getString(R.string.preference_server_address), "");
		String serverPort = prefs.getString(getString(R.string.preference_server_port), "");
		restClient = new RestClient(serverAddress + ":" + serverPort);

		if (selectedChannel.size() > 1) {
			showChannelDetails(true);
			boolean alwaysShowChannels = prefs.getBoolean(getString(R.string.pref_key_showChannels), false);
			if (getResources().getConfiguration().orientation == Configuration.ORIENTATION_PORTRAIT
					&& !alwaysShowChannels) {
				findViewById(R.id.channelListFragment).setVisibility(View.GONE);
			}
			timeSpanChanged();
		}

	}

	@Override
	protected void onResume() {
		super.onResume();

		SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(this);
		int interval = Integer.parseInt(prefs.getString(getString(R.string.pref_key_update_interval), "0"));
		updater.start(listFragment.getAdapterHandler(), restClient, interval);
	}

	@Override
	protected void onPause() {
		super.onPause();

		updater.stop();
	}

	@Override
	protected void onStop() {
		super.onStop();

		updater.setCallback(null);
		restClient.release();
	}

	@Override
	protected void onSaveInstanceState(Bundle outState) {
		super.onSaveInstanceState(outState);

		outState.putInt("seriesNumber", channelSeries.size());
		int i = 0;

		Iterator<ChannelRecordSeries> iter = channelSeries.values().iterator();
		while (iter.hasNext()) {
			ChannelRecordSeries series = iter.next();
			List<Record> records = series.slice(series.getEarliestTimestamp() - 1, series.getLatestTimestamp() + 1);
			outState.putString("seriesLabel_" + i, series.getTitle());
			outState.putParcelableArray("seriesRecords_" + i, records.toArray(new Record[0]));
			i++;
		}
	}

	@Override
	public boolean onCreateOptionsMenu(Menu menu) {
		// Inflate the menu; this adds items to the action bar if it is present.
		getMenuInflater().inflate(R.menu.main, menu);
		return true;
	}

	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
		switch (item.getItemId()) {
		case android.R.id.home:
			// This ID represents the Home or Up button. In the case of this
			// activity, the Up button is shown. Use NavUtils to allow users
			// to navigate up one level in the application structure. For
			// more details, see the Navigation pattern on Android Design:
			//
			// http://developer.android.com/design/patterns/navigation.html#up-vs-back
			//
			NavUtils.navigateUpFromSameTask(this);
			return true;
		case R.id.action_settings:
			Intent goToSettings = new Intent(this, SettingsActivity.class);
			startActivity(goToSettings);
			return true;
		case R.id.action_addchannel:
			Intent goToAddChannels = new Intent(this, AddChannelActivity.class);
			startActivityForResult(goToAddChannels, AddChannelActivity.REQUEST_CODE);
			return true;
		case R.id.action_refresh:
			updater.pollUpdate();
			return true;
		}
		return super.onOptionsItemSelected(item);
	}

	@Override
	protected void onActivityResult(int requestCode, int resultCode, Intent data) {
		if (requestCode == AddChannelActivity.REQUEST_CODE) {
			if (resultCode == RESULT_OK) {
				String[] newChannels = data.getStringArrayExtra("channelLabels");
				Set<String> storedChannels = SharedPreferenceUtil.getChannelLabels(this);
				Collections.addAll(storedChannels, newChannels);
				SharedPreferenceUtil.writeChannelLabels(this, storedChannels);

				synchronized (channelsToRead) {
					for (String label : newChannels) {
						Channel newChannel = new Channel();
						newChannel.setLabel(label);
						channelsToRead.add(newChannel);
					}
				}
			}
		}
	}

	@Override
	public void channelSelected(Channel channel) {
		if (selectedChannel.contains(channel)) {
			selectedChannel.remove(channel);
			channelSeries.remove(channel);
			if (channelSeries.size() == 0) {
				listFragment.getListView().setSelected(false);
				showChannelDetails(false);
				findViewById(R.id.channelListFragment).setVisibility(View.VISIBLE);
			}
			else {
				timeSpanChanged();
			}
		}
		else if (channelSeries.size() < plotFragment.getMaxChannels()) {
			selectedChannel.add(channel);
			List<Record> dummy = Collections.emptyList();
			channelSeries.put(channel, new ChannelRecordSeries(channel.getLabel(), null, dummy));
			if (selectedChannel.size() == 1) {
				showChannelDetails(true);
				SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(this);
				boolean alwaysShowChannels = prefs.getBoolean(getString(R.string.pref_key_showChannels), false);
				if (getResources().getConfiguration().orientation == Configuration.ORIENTATION_PORTRAIT
						&& !alwaysShowChannels) {
					findViewById(R.id.channelListFragment).setVisibility(View.GONE);
				}
			}
			timeSpanChanged();
		}
	}

	@Override
	public void deleteChannel(Set<String> channelsToDelete) {
		updater.stop();

		Iterator<Channel> iter = selectedChannel.iterator();
		while (iter.hasNext()) {
			Channel c = iter.next();
			if (channelsToDelete.contains(c.getLabel())) {
				if (selectedChannel.size() == 1) {
					listFragment.getListView().setSelected(false);
					showChannelDetails(false);
				}
				iter.remove();
				break;
			}
		}

		Set<String> channels = SharedPreferenceUtil.getChannelLabels(this);
		channels.removeAll(channelsToDelete);
		SharedPreferenceUtil.writeChannelLabels(this, channels);

		iter = channelsToRead.iterator();
		while (iter.hasNext()) {
			Channel channel = iter.next();
			if (channelsToDelete.contains(channel.getLabel())) {
				iter.remove();
			}
		}

		SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(this);
		int interval = Integer.parseInt(prefs.getString(getString(R.string.pref_key_update_interval), "0"));
		updater.start(listFragment.getAdapterHandler(), restClient, interval);
	}

	@Override
	public void timeSpanChanged() {
		if (selectedChannel != null) {
			updater.setCallback(null);

			LoadHistoryTask task;
			Long duration = timespanFragment.getDuration();
			if (duration == null || duration <= 0) {
				task = new LoadHistoryTask(restClient, timespanFragment.getStartTimestamp(),
						timespanFragment.getEndTimestamp());
			}
			else {
				task = new LoadHistoryTask(restClient, duration);
			}

			plotFragment.showPlot(false);
			task.execute(selectedChannel.toArray(new Channel[0]));
		}
	}

	/**
	 * Enables or disables the graph for selected channels
	 * 
	 * @param show
	 */
	public void showChannelDetails(boolean show) {
		View detailView = findViewById(R.id.channelDetailsLayout);
		int visibility = show ? View.VISIBLE : View.GONE;
		detailView.setVisibility(visibility);
	}

	/**
	 * AsyncTask used to load the history of the selected channel at the given time frame. If the task was successful,
	 * the graph of the channel will be displayed at the device and, if set by the user, continuously updated.
	 * 
	 * @author Karsten Müller-Bier
	 */
	private class LoadHistoryTask extends AsyncTask<Channel, Void, Void> {

		private final RestClient client;
		private final Long maxDuration;
		private final Long from;
		private final Long until;

		public LoadHistoryTask(RestClient client, Long maxDuration) {
			this.client = client;
			this.maxDuration = maxDuration;
			until = System.currentTimeMillis();
			from = until - maxDuration;
		}

		public LoadHistoryTask(RestClient client, Long from, Long until) {
			this.client = client;
			maxDuration = null;
			this.from = from;
			this.until = until;
		}

		@Override
		protected Void doInBackground(Channel... params) {
			if (params.length == 0) {
				cancel(true);
				return null;
			}

			for (Channel channel : params) {
				ChannelRecordSeries series = channelSeries.get(channel);

				List<Record> records = new LinkedList<Record>();
				List<Record> middle = series.slice(from, until);
				List<Record> begin = Collections.emptyList();
				List<Record> end = Collections.emptyList();

				try {
					if (from < series.getEarliestTimestamp()) {
						long untilMiddle = Math.min(until, series.getEarliestTimestamp() - 1);
						begin = client.readChannelHistory(channel, from, untilMiddle);
					}
					if (until > series.getLatestTimestamp()) {
						long fromMiddle = Math.max(from, series.getLatestTimestamp() + 1);
						end = client.readChannelHistory(channel, fromMiddle, until);
					}
				} catch (IOException e) {
					cancel(true);
					return null;
				}

				records.addAll(begin);
				records.addAll(middle);
				records.addAll(end);
				channelSeries.put(channel, new ChannelRecordSeries(channel.getLabel(), maxDuration, records));
			}

			return null;
		}

		@Override
		protected void onPostExecute(Void result) {

			plotFragment.clearPlotSeries();
			int i = 0;
			for (Channel key : channelSeries.keySet()) {
				plotFragment.setPlotSeries(channelSeries.get(key), i++);
			}
			plotFragment.showPlot(true);

			RelativeLayout layout = (RelativeLayout) findViewById(R.id.chartLayout);
			layout.invalidate();

			if (maxDuration != null) {
				updater.setCallback(DashBoard.this);
			}
		}
	}

	@Override
	public void onBackPressed() {
		if (!selectedChannel.isEmpty()) {
			listFragment.getListView().setSelected(false);
			selectedChannel.clear();
			channelSeries.clear();
			showChannelDetails(false);
			if (getResources().getConfiguration().orientation == Configuration.ORIENTATION_PORTRAIT) {
				findViewById(R.id.channelListFragment).setVisibility(View.VISIBLE);
			}
		}
		else {
			super.onBackPressed();
		}
	}

	@Override
	public void channelUpdated(Channel c) {
		channelSeries.get(c).addRecord(c.getRecord());
		newRecord = true;
	}

	@Override
	public Collection<Channel> getListenedChannels() {
		return channelSeries.keySet();
	}

	@Override
	public void cycleFinished() {
		if (newRecord) {
			plotFragment.replot();
			newRecord = false;
		}

	}
}
