package org.openmuc.android.fragments;

import java.text.DecimalFormat;
import java.text.FieldPosition;
import java.text.Format;
import java.text.ParsePosition;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.openmuc.android.R;
import org.openmuc.android.util.ChannelRecordSeries;

import android.content.SharedPreferences;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.os.Build;
import android.os.Bundle;
import android.preference.PreferenceManager;
import android.support.v4.app.Fragment;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;

import com.androidplot.Plot;
import com.androidplot.PlotEvent;
import com.androidplot.PlotListener;
import com.androidplot.ui.AnchorPosition;
import com.androidplot.ui.SizeLayoutType;
import com.androidplot.ui.SizeMetrics;
import com.androidplot.xy.BoundaryMode;
import com.androidplot.xy.FillDirection;
import com.androidplot.xy.LineAndPointFormatter;
import com.androidplot.xy.XLayoutStyle;
import com.androidplot.xy.XPositionMetric;
import com.androidplot.xy.XYPlot;
import com.androidplot.xy.XYStepMode;
import com.androidplot.xy.YLayoutStyle;
import com.androidplot.xy.YValueMarker;

// Suppressing the deprecation warning of the PlotEvent class. That class is part of the PlotListener interface and 
// cannot be removed without removing the whole interface
@SuppressWarnings("deprecation")
/**
 * Fragment that displays the channel history for up to 3 channels provided by the activity with setPlotSeries.
 * 
 * @author Karsten Müller-Bier
 */
public class ChannelChartFragment extends Fragment implements PlotListener {

	private static final int LISTSIZE = 4;

	private boolean isDrawing = false;

	private XYPlot dynamicPlot;
	private final List<LineAndPointFormatter> formatterList = new ArrayList<LineAndPointFormatter>(LISTSIZE);
	private final List<ChannelRecordSeries> seriesList = new ArrayList<ChannelRecordSeries>(LISTSIZE);

	/**
	 * @return The maximum amount of channels that can be displayed at a given time
	 */
	public int getMaxChannels() {
		return LISTSIZE;
	}

	/**
	 * Orders the history graph to redraw itself. If the graph is currently redrawing itself, nothing happens.
	 */
	public void replot() {
		if (!isDrawing) {
			long min = 0;
			double yMin = 0, yMax = 0;
			for (ChannelRecordSeries series : seriesList) {
				min = Math.max(min, series.getX(0).longValue());
				yMin = Math.min(yMin, Math.abs(series.getMin()));
				yMax = Math.max(yMax, Math.abs(series.getMax()));
			}
			StringBuilder sb = new StringBuilder("1.");
			if (yMin < 1000 || yMax < 1000) {
				sb.append("#");
			}
			if (yMin < 100 || yMax < 100) {
				sb.append("#");
			}
			if (yMin < 10 || yMax < 10) {
				sb.append("#");
			}
			dynamicPlot.setRangeValueFormat(new DecimalFormat(sb.toString()));
			dynamicPlot.setDomainLowerBoundary(min, BoundaryMode.FIXED);
			dynamicPlot.redraw();
		}
	}

	/**
	 * Toggles between showing the graph or the loading circle. The loading circle should be shown while the history of
	 * a channel is being loaded.
	 * 
	 * @param show
	 *            Setting to true will show the graph, false shows the loading circle instead
	 */
	public void showPlot(boolean show) {
		View progressBar = getActivity().findViewById(R.id.chartProgress);
		progressBar.setVisibility(show ? View.INVISIBLE : View.VISIBLE);

		View plot = getActivity().findViewById(R.id.plot);
		plot.setVisibility(show ? View.VISIBLE : View.INVISIBLE);
	}

	/**
	 * Adds an additional series to the graph
	 * 
	 * @param series
	 *            reference of the series to add
	 * @param index
	 *            position of the series
	 * @throws IndexOutOfBoundsException
	 *             Thrown if index is greater or equal getMaxChannels()
	 */
	public void setPlotSeries(ChannelRecordSeries series, int index) {
		index = Math.max(index, seriesList.size());

		seriesList.add(index, series);

		if (seriesList.size() <= 4) {
			for (LineAndPointFormatter formatter : formatterList) {
				formatter.getFillPaint().setAlpha(80 - 20 * seriesList.size());
			}
		}

		dynamicPlot.clear();
		for (int i = 0; i < seriesList.size(); i++) {
			dynamicPlot.addSeries(seriesList.get(i), formatterList.get(i));
		}
	}

	/**
	 * Removes all series' from the graph
	 */
	public void clearPlotSeries() {
		dynamicPlot.clear();
		seriesList.clear();
	}

	@Override
	public View onCreateView(LayoutInflater inflater, ViewGroup container, Bundle savedInstanceState) {
		return inflater.inflate(R.layout.fragment_channel_chart, container, false);
	}

	@Override
	public void onStart() {
		super.onStart();

		dynamicPlot = (XYPlot) getActivity().findViewById(R.id.plot);
		dynamicPlot.setRangeValueFormat(new DecimalFormat("0"));
		dynamicPlot.getGraphWidget().setMarginTop(10);
		dynamicPlot.getGraphWidget().setRangeLabelWidth(50);
		dynamicPlot.setDomainStepMode(XYStepMode.INCREMENT_BY_PIXELS);
		dynamicPlot.setDomainStepValue(120);
		dynamicPlot.setDomainValueFormat(new Format() {
			private static final long serialVersionUID = 1L;
			// create a simple date format that draws on the year portion of our timestamp.
			// see http://download.oracle.com/javase/1.4.2/docs/api/java/text/SimpleDateFormat.html
			// for a full description of SimpleDateFormat.
			private final SimpleDateFormat dateFormat = new SimpleDateFormat("hh:mm:ss", Locale.US);

			@Override
			public StringBuffer format(Object obj, StringBuffer toAppendTo, FieldPosition pos) {
				Date date = new Date(((Number) obj).longValue());
				return dateFormat.format(date, toAppendTo, pos);
			}

			@Override
			public Object parseObject(String source, ParsePosition pos) {
				return null;

			}
		});

		Paint marker = new Paint();
		marker.setStrokeWidth(2);
		if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.HONEYCOMB) {
			marker.setColor(Color.BLACK);
			dynamicPlot.getGraphWidget().getBackgroundPaint().setColor(Color.WHITE);
			dynamicPlot.getGraphWidget().getGridBackgroundPaint().setColor(Color.WHITE);
			dynamicPlot.getGraphWidget().getDomainLabelPaint().setColor(Color.BLACK);
			dynamicPlot.getGraphWidget().getRangeLabelPaint().setColor(Color.BLACK);
			dynamicPlot.getGraphWidget().getDomainOriginLabelPaint().setColor(Color.BLACK);
			dynamicPlot.getGraphWidget().getDomainOriginLinePaint().setColor(Color.BLACK);
			dynamicPlot.getGraphWidget().getRangeOriginLinePaint().setColor(Color.BLACK);

			Paint background = new Paint();
			background.setColor(Color.WHITE);
			dynamicPlot.setBackgroundPaint(background);
		}
		else {
			marker.setColor(Color.LTGRAY);
		}

		if (isLargeTablet()) {
			dynamicPlot.getGraphWidget().setMarginBottom(25);
			dynamicPlot.getLegendWidget().setHeight(35);
			dynamicPlot.getLegendWidget().setWidth(1);
			dynamicPlot.getLegendWidget().setPaddingLeft(5);
			dynamicPlot.position(dynamicPlot.getLegendWidget(), 0, XLayoutStyle.ABSOLUTE_FROM_LEFT, 0,
					YLayoutStyle.ABSOLUTE_FROM_BOTTOM, AnchorPosition.LEFT_BOTTOM);
			dynamicPlot.getLegendWidget().getTextPaint().setColor(Color.BLACK);
			dynamicPlot.getLegendWidget().getTextPaint().setTextSize(20);
			dynamicPlot.getLegendWidget().setIconSizeMetrics(
					new SizeMetrics(20, SizeLayoutType.ABSOLUTE, 20, SizeLayoutType.ABSOLUTE));
		}
		else {
			dynamicPlot.getGraphWidget().setMarginBottom(0);
			dynamicPlot.getLayoutManager().remove(dynamicPlot.getLegendWidget());
		}

		dynamicPlot.getLayoutManager().remove(dynamicPlot.getDomainLabelWidget());
		dynamicPlot.getLayoutManager().remove(dynamicPlot.getRangeLabelWidget());
		dynamicPlot.getLayoutManager().remove(dynamicPlot.getTitleWidget());

		SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(getActivity());
		String minRange = prefs.getString(getString(R.string.pref_key_y_axis_min), "");
		String maxRange = prefs.getString(getString(R.string.pref_key_y_axis_max), "");
		if (minRange != "") {
			try {
				Double min = Double.parseDouble(minRange);
				dynamicPlot.setRangeLowerBoundary(min, BoundaryMode.FIXED);
			} catch (NumberFormatException ex) {
			}
		}
		else {
			dynamicPlot.setRangeLowerBoundary(null, BoundaryMode.AUTO);
		}
		if (maxRange != "") {
			try {
				Double max = Double.parseDouble(maxRange);
				dynamicPlot.setRangeUpperBoundary(max, BoundaryMode.FIXED);
			} catch (NumberFormatException ex) {
			}
		}
		else {
			dynamicPlot.setRangeUpperBoundary(null, BoundaryMode.AUTO);
		}

		dynamicPlot.addMarker(new YValueMarker(0, "", new XPositionMetric(0, XLayoutStyle.ABSOLUTE_FROM_LEFT), marker,
				marker));

		LineAndPointFormatter f1 = new LineAndPointFormatter(Color.rgb(0, 100, 0), null, Color.rgb(0, 80, 0), null,
				FillDirection.BOTTOM);
		LineAndPointFormatter f2 = new LineAndPointFormatter(Color.rgb(200, 0, 0), null, Color.rgb(150, 0, 0), null,
				FillDirection.BOTTOM);
		LineAndPointFormatter f3 = new LineAndPointFormatter(Color.rgb(0, 0, 200), null, Color.rgb(0, 0, 150), null,
				FillDirection.BOTTOM);
		LineAndPointFormatter f4 = new LineAndPointFormatter(Color.rgb(180, 180, 0), null, Color.rgb(80, 80, 0), null,
				FillDirection.BOTTOM);
		f1.getLinePaint().setStrokeWidth(2);
		f2.getLinePaint().setStrokeWidth(2);
		f3.getLinePaint().setStrokeWidth(2);
		f4.getLinePaint().setStrokeWidth(2);
		f1.getFillPaint().setAlpha(60);
		f2.getFillPaint().setAlpha(60);
		f3.getFillPaint().setAlpha(60);
		f4.getFillPaint().setAlpha(60);

		formatterList.add(f1);
		formatterList.add(f2);
		formatterList.add(f3);
		formatterList.add(f4);
	}

	@SuppressWarnings("rawtypes")
	@Override
	public void onAfterDraw(Plot arg0, Canvas arg1) {
		isDrawing = false;
	}

	@SuppressWarnings("rawtypes")
	@Override
	public void onBeforeDraw(Plot arg0, Canvas arg1) {
		isDrawing = true;
	}

	@Override
	public void onPlotUpdate(PlotEvent arg0) {
		// ignore, method is deprecated
	}

	private boolean isLargeTablet() {
		int width = getResources().getDisplayMetrics().widthPixels;
		int height = getResources().getDisplayMetrics().heightPixels;
		int min = Math.min(width, height);

		return min >= 600;
	}
}
