package org.openmuc.android.fragments;

import java.util.Calendar;

import org.openmuc.android.R;
import org.openmuc.android.util.TimeSpan;

import android.app.Activity;
import android.os.Bundle;
import android.support.v4.app.Fragment;
import android.text.format.DateFormat;
import android.view.LayoutInflater;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.ViewGroup;
import android.widget.AdapterView;
import android.widget.AdapterView.OnItemSelectedListener;
import android.widget.ArrayAdapter;
import android.widget.LinearLayout;
import android.widget.Spinner;
import android.widget.TextView;

import com.ptashek.widgets.datetimepicker.DateTimePickerDialog;
import com.ptashek.widgets.datetimepicker.DateTimePickerDialog.OnSubmitListener;

/**
 * Fragment that displays all layout elements to control the time span of the graph. <br>
 * <br>
 * An activity that uses this Fragment must implement the {@link ChartCallback} interface
 * 
 * @author Karsten Müller-Bier
 */
public class TimespanSelectorFragment extends Fragment implements OnClickListener {

	private java.text.DateFormat dateFormat;
	private java.text.DateFormat timeFormat;

	private Calendar startDate = Calendar.getInstance();
	private Calendar endDate = Calendar.getInstance();
	private Long duration;

	private ChartCallback callback;

	private DateTimePickerDialog dateTimeDialog;

	private Integer spinnerSelection = null;

	/**
	 * @return The starting time in milliseconds since UNIX EPOCH
	 */
	public long getStartTimestamp() {
		return startDate.getTimeInMillis();
	}

	/**
	 * @return The ending time in milliseconds since UNIX EPOCH
	 */
	public long getEndTimestamp() {
		return endDate.getTimeInMillis();
	}

	/**
	 * @return The duration the dynamic graph shall display in milliseconds
	 */
	public Long getDuration() {
		return duration;
	}

	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);

		dateFormat = DateFormat.getDateFormat(getActivity());
		timeFormat = DateFormat.getTimeFormat(getActivity());
		dateTimeDialog = new DateTimePickerDialog(getActivity());

		if (savedInstanceState != null) {
			spinnerSelection = savedInstanceState.getInt("spinnerKey");
		}
	}

	@Override
	public View onCreateView(LayoutInflater inflater, ViewGroup container, Bundle savedInstanceState) {
		return inflater.inflate(R.layout.fragment_timespan_seletor, container, false);
	}

	@Override
	public void onStart() {
		super.onStart();

		final View startLayout = getActivity().findViewById(R.id.fromSelectorLayout);
		final View endLayout = getActivity().findViewById(R.id.untilSelectorLayout);

		Spinner timeSpanSpinner = (Spinner) getActivity().findViewById(R.id.timeSpanSpinner);
		timeSpanSpinner.setAdapter(new ArrayAdapter<TimeSpan>(getActivity(), R.layout.single_row, TimeSpan.values()));
		timeSpanSpinner.setOnItemSelectedListener(new OnItemSelectedListener() {

			@Override
			public void onItemSelected(AdapterView<?> spinner, View view, int position, long id) {
				TimeSpan selected = (TimeSpan) spinner.getAdapter().getItem(position);

				if (selected.getValue() != -1) {
					startLayout.setVisibility(View.GONE);
					endLayout.setVisibility(View.GONE);
					duration = (long) selected.getValue() * 1000;
					endDate.setTimeInMillis(System.currentTimeMillis());
					startDate.setTimeInMillis(endDate.getTimeInMillis() - duration);
					callback.timeSpanChanged();
				}
				else {
					startLayout.setVisibility(View.VISIBLE);
					endLayout.setVisibility(View.VISIBLE);
					duration = null;
					TextView startText = (TextView) ((LinearLayout) startLayout).getChildAt(1);
					startText.setText(dateFormat.format(startDate.getTime()) + " "
							+ timeFormat.format(startDate.getTime()));
					TextView endText = (TextView) ((LinearLayout) endLayout).getChildAt(1);
					endText.setText(dateFormat.format(endDate.getTime()) + " " + timeFormat.format(endDate.getTime()));
				}
			}

			@Override
			public void onNothingSelected(AdapterView<?> spinner) {
				spinner.setSelection(0);
			}

		});

		if (spinnerSelection != null) {
			timeSpanSpinner.setSelection(spinnerSelection);
			spinnerSelection = null;
		}
		else {
			timeSpanSpinner.setSelection(TimeSpan._60.ordinal());
		}

		getActivity().findViewById(R.id.startDateButton).setOnClickListener(this);
		getActivity().findViewById(R.id.endDateButton).setOnClickListener(this);
	}

	@Override
	public void onAttach(Activity activity) {
		super.onAttach(activity);

		if (activity instanceof ChartCallback) {
			callback = (ChartCallback) activity;
		}
		else {
			throw new IllegalArgumentException(activity.getClass().getName() + " does not implement "
					+ ChartCallback.class.getName());
		}
	}

	@Override
	public void onDetach() {
		super.onDetach();
		callback = null;
	}

	@Override
	public void onSaveInstanceState(Bundle outState) {
		super.onSaveInstanceState(outState);
		Spinner timeSpanSpinner = (Spinner) getActivity().findViewById(R.id.timeSpanSpinner);
		outState.putInt("spinnerKey", timeSpanSpinner.getSelectedItemPosition());
	}

	@Override
	public void onClick(View view) {
		if (view.getId() == R.id.startDateButton) {
			dateTimeDialog.showDateTimeDialog(new OnSubmitListener() {
				@Override
				public void submit(Calendar result) {
					if (result.before(endDate)) {
						startDate = result;
						TextView startText = (TextView) getActivity().findViewById(R.id.startDateText);
						startText.setText(dateFormat.format(startDate.getTime()) + " "
								+ timeFormat.format(startDate.getTime()));

						callback.timeSpanChanged();
					}
				}
			});
		}
		else if (view.getId() == R.id.endDateButton) {
			dateTimeDialog.showDateTimeDialog(new OnSubmitListener() {
				@Override
				public void submit(Calendar result) {
					if (result.after(startDate)) {
						endDate = result;
						TextView endText = (TextView) getActivity().findViewById(R.id.endDateText);
						endText.setText(dateFormat.format(endDate.getTime()) + " "
								+ timeFormat.format(endDate.getTime()));

						callback.timeSpanChanged();
					}
				}
			});
		}
	}

	/**
	 * Interface used by {@link TimespanSelectorFragment} to notify its activity about user interaction
	 * 
	 * @author Karsten Müller-Bier
	 */
	public interface ChartCallback {
		/**
		 * Callback to notify the activity that either startTime, endTime or duration have been changed
		 */
		void timeSpanChanged();
	}
}
