package org.openmuc.android.rest;

import java.io.IOException;
import java.util.Collections;
import java.util.List;
import java.util.logging.Level;

import org.openmuc.android.rest.data.Channel;
import org.openmuc.android.rest.data.Record;
import org.restlet.Client;
import org.restlet.Request;
import org.restlet.Response;
import org.restlet.data.Method;
import org.restlet.data.Protocol;
import org.restlet.data.Reference;
import org.restlet.resource.ClientResource;

import android.util.Log;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;

/**
 * Client-side endpoint of a RESTful communication.
 * 
 * A RestClient manages all objects needed to communicate with the RESTful Webservice defined in the
 * openmuc-server-restws project
 * 
 * @author Karsten Müller-Bier
 */
public class RestClient {

	private final ObjectMapper mapper;
	private final Client client;
	private final Reference serverRef;

	/**
	 * @param server
	 *            The URL of the server
	 * @param useHttps
	 */
	public RestClient(String server) {
		server = fixServerString(server);

		// Initialize Jackson
		mapper = new ObjectMapper();

		// Initialize Restlet Client
		serverRef = new Reference(server).addSegment("rest").addSegment("channel").addSegment("");
		client = new Client(Protocol.HTTP);
		client.getLogger().setLevel(Level.WARNING);
	}

	public void release() {
		try {
			client.stop();
		} catch (Exception e) {
			Log.d("RestClient", e.getMessage());
		}
	}

	/**
	 * Adds the right protocol prefix to the server URL, if it's missing
	 * 
	 * @param server
	 *            The URL of the server
	 * @param useHttps
	 * @return The fixed Address
	 */
	private String fixServerString(String server) {
		if (!server.startsWith("http://")) {
			return "http://" + server;
		}
		return server;
	}

	/**
	 * Retrieves a list of all registered channels on the server. The list is deserialized using a TypeReference object.
	 * These objects are needed if the JSON Stream has to be deserialized into a generic class.
	 * 
	 * @return The list of remote channels
	 * @throws IOException
	 */
	public List<Channel> listChannels() throws IOException {
		serverRef.setLastSegment("");
		Request req = new Request(Method.GET, serverRef);
		Response resp = client.handle(req);

		List<Channel> channels = Collections.emptyList();
		if (resp.getStatus().isSuccess()) {
			channels = mapper.readValue(resp.getEntity().getStream(), new TypeReference<List<Channel>>() {
			});
		}
		resp.release();
		return channels;
	}

	public List<Record> readChannelHistory(Channel channel, long start, long end) throws IOException {
		serverRef.setLastSegment(channel.getLabel());
		Reference historyRef = new Reference(serverRef);
		historyRef.addSegment("history");
		historyRef.addQueryParameter("from", "" + start);
		historyRef.addQueryParameter("until", "" + end);

		Request req = new Request(Method.GET, historyRef);

		Response resp = client.handle(req);

		List<Record> result = null;
		if (resp.getStatus().isSuccess()) {
			result = mapper.readValue(resp.getEntity().getStream(), new TypeReference<List<Record>>() {
			});
		}
		else {
			result = Collections.emptyList();
		}
		resp.release();
		return result;
	}

	/**
	 * Retrieves the latest record of a channel. The record is deserialized using the class as reference. For
	 * deserializing Generics, see {@link RestClient#listChannels()}.
	 * 
	 * To read the latest record, a copy of the ClientResource instance variable is copied first and the label of the
	 * channel to read is added to the copy. After calling GET and retrieving the record, the copied ClientResourced is
	 * cleaned up using {@link ClientResource#release()}.
	 * 
	 * @param channel
	 *            The channel to read
	 * @return true if the received sample is newer than the previous one
	 * @throws IOException
	 */
	public boolean readChannel(Channel channel) throws IOException {
		serverRef.setLastSegment(channel.getLabel());
		Request req = new Request(Method.GET, serverRef);
		Response resp = client.handle(req);

		if (resp.getStatus().isSuccess()) {
			Record newRecord = mapper.readValue(resp.getEntity().getStream(), Record.class);
			if (channel.getRecord().getTimestamp() < newRecord.getTimestamp()) {
				channel.setRecord(newRecord);
				return true;
			}
		}
		else {
			if (resp.getStatus().isClientError()) {
				channel.getRecord().setFlag("Channel not found");
			}
			else if (resp.getStatus().isServerError()) {
				channel.getRecord().setFlag("Server error");
			}
			else {
				channel.getRecord().setFlag("No connection to server");
			}
		}
		resp.release();
		return false;
	}
}
