package org.openmuc.android.util;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;

import org.openmuc.android.fragments.ChannelChartFragment;
import org.openmuc.android.rest.data.Record;

import com.androidplot.series.XYSeries;

/**
 * Container used by {@link ChannelChartFragment} that holds all records of a time frame
 * 
 * @author Karsten Müller-Bier
 */
public class ChannelRecordSeries implements XYSeries {

	private List<Record> dataset;
	private final String title;
	private final Long maxDuration;

	private double max = 0;
	private double min = 0;

	public ChannelRecordSeries(String title, Long maxDuration, List<Record> records) {
		this.title = title;
		this.maxDuration = maxDuration;
		if (maxDuration != null) {
			this.dataset = new ArrayList<Record>((int) (maxDuration / 10) + 1);
		}
		else {
			this.dataset = new ArrayList<Record>(records.size() + 1);
		}
		for (Record r : records) {
			if ("VALID".equalsIgnoreCase(r.getFlag())) {
				dataset.add(r);
				Math.max(max, r.getValue());
				Math.min(min, r.getValue());
			}
		}
	}

	@Override
	public String getTitle() {
		return title;
	}

	@Override
	public int size() {
		if (dataset.size() > 0) {
			return dataset.size();
		}
		return 1;
	}

	@Override
	public Number getX(int index) {
		if (dataset.size() == 0 || index >= dataset.size()) {
			return 0.0;
		}
		return dataset.get(index).getTimestamp();
	}

	@Override
	public Number getY(int index) {
		if (dataset.size() == 0 || index >= dataset.size()) {
			return 0.0;
		}
		return dataset.get(index).getValue();
	}

	/**
	 * Adds a record to the series. If a maxDuration has been set during construction, all records older than the
	 * timestamp of the new record - maxDuration are removed
	 * 
	 * @param r
	 *            the new Record
	 */
	public void addRecord(Record r) {
		dataset.add(r);
		Math.max(max, r.getValue());
		Math.min(min, r.getValue());
		if (maxDuration != null) {
			long timegap = 0;
			if (dataset.size() > 1) {
				Record previous = dataset.get(dataset.size() - 2);
				timegap = r.getTimestamp() - previous.getTimestamp();
			}
			long timestamp = r.getTimestamp() - maxDuration - timegap;
			Iterator<Record> iter = dataset.iterator();
			while (iter.hasNext()) {
				Record oldRecord = iter.next();
				if (oldRecord.getTimestamp() < timestamp) {
					iter.remove();
				}
				else {
					break;
				}
			}
		}
	}

	/**
	 * @return The timestamp of the oldest record, or 0 if no record exists
	 */
	public long getEarliestTimestamp() {
		return getX(0).longValue();
	}

	/**
	 * @return The timestamp of the newest record, or 0 if no record exists
	 */
	public long getLatestTimestamp() {
		return getX(dataset.size() - 1).longValue();
	}

	/**
	 * Removes all records whose timestamp is not in the range between begin and end
	 * 
	 * @param begin
	 *            Timestamp of the oldest record allowed in milliseconds since UNIX EPOCH
	 * @param end
	 *            Timestamp of the newest record allowed in milliseconds since UNIX EPOCH
	 * @return List of all records in range
	 */
	public List<Record> slice(long begin, long end) {
		long earliest = getEarliestTimestamp();
		long latest = getLatestTimestamp();
		if (end <= begin || end <= earliest || begin >= latest) {
			dataset.clear();
			return dataset;
		}
		if (begin < earliest && end > latest) {
			return dataset;
		}

		List<Record> toRemove = new LinkedList<Record>();

		ListIterator<Record> iter = dataset.listIterator();
		while (iter.hasNext()) {
			Record record = iter.next();
			if (record.getTimestamp() < begin) {
				toRemove.add(record);
			}
			else {
				break;
			}
		}

		iter = dataset.listIterator(dataset.size() - 1);
		while (iter.hasPrevious()) {
			Record record = iter.previous();
			if (record.getTimestamp() > end) {
				toRemove.add(record);
			}
			else {
				break;
			}
		}

		dataset.removeAll(toRemove);
		return dataset;
	}

	public double getMax() {
		return max;
	}

	public double getMin() {
		return min;
	}
}
