package org.openmuc.android.util;

import java.io.IOException;

import org.openmuc.android.rest.RestClient;
import org.openmuc.android.rest.data.Channel;
import org.openmuc.android.rest.data.Record;

/**
 * Runner class that updates a single channel in a given interval. Additionally features a callback to be notified when
 * the channel received a new record. <br>
 * An ChannelRecordUpdater instance can be started several times using its start() method
 * 
 * @author Karsten Müller-Bier
 */
public class ChannelRecordUpdater implements Runnable {

	private final RestClient client;
	private Channel channel;
	private RecordUpdaterCallback callback;

	private boolean running = false;
	private int interval;
	private Thread activeThread;

	public ChannelRecordUpdater(RestClient client) {
		this.client = client;
	}

	/**
	 * Starts the runner. Does nothing if the runner is still alive
	 * 
	 * @param channel
	 *            The channel to read periodically
	 * @param callback
	 * @param interval
	 *            time between two reads in milliseconds
	 */
	public void start(Channel channel, RecordUpdaterCallback callback, int interval) {
		this.channel = channel;
		this.interval = interval;
		this.callback = callback;

		new Thread() {
			@Override
			public void run() {
				if (activeThread != null && activeThread.isAlive()) {
					if (!activeThread.isInterrupted()) {
						return;
					}
					try {
						activeThread.join();
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
				running = true;
				activeThread = new Thread(ChannelRecordUpdater.this);
				activeThread.setDaemon(true);
				activeThread.setName("ChannelRecordUpdater");
				activeThread.start();
			};
		}.run();
	}

	/**
	 * Calls the runner to stop at the next opportunity
	 */
	public void stop() {
		if (activeThread != null) {
			running = false;
			activeThread.interrupt();
			activeThread = null;
		}
	}

	@Override
	public void run() {
		boolean newValue = false;
		while (running) {
			try {
				newValue = client.readChannel(channel);
			} catch (IOException e) {
				newValue = false;
			}
			if (newValue) {
				callback.newRecord(channel.getRecord());
			}
			try {
				Thread.sleep(interval);
			} catch (InterruptedException e) {
				running = false;
			}
		}
	}

	/**
	 * Interface used by {@link ChannelRecordUpdater} to notify that a new Record has been received
	 * 
	 * @author Karsten Müller-Bier
	 */
	public interface RecordUpdaterCallback {
		/**
		 * @param record
		 *            the received record
		 */
		void newRecord(Record record);
	}
}
