package org.openmuc.android.util;

import java.io.IOException;
import java.util.Collection;
import java.util.List;

import org.openmuc.android.rest.RestClient;
import org.openmuc.android.rest.data.Channel;

import android.os.Handler;
import android.util.Log;

/**
 * Runner class that updates all channels provided by a list in a given interval. Additionally features a callback to be
 * notified when a channel received a new record. <br>
 * An ChannelUpdater instance can be started several times using its start() method
 * 
 * @author Karsten Müller-Bier
 */
public class ChannelUpdater implements Runnable {

	private Thread runningThread;
	private final List<Channel> channels;
	private RestClient rest;

	private boolean running;
	private int interval;
	private Handler handler;

	private UpdaterCallback callback;

	public ChannelUpdater(List<Channel> channels) {
		this.channels = channels;
	}

	public void setCallback(UpdaterCallback callback) {
		this.callback = callback;
	}

	@Override
	public void run() {
		long start = System.currentTimeMillis();
		long nextStart = start - start % interval;
		while (running) {
			try {
				start = System.currentTimeMillis();
				while (nextStart <= start) {
					nextStart += interval;
				}
				long diff = nextStart - start;
				Thread.sleep(diff);
				synchronized (channels) {
					for (Channel channel : channels) {
						try {
							boolean updated = rest.readChannel(channel);
							if (callback != null && updated && callback.getListenedChannels().contains(channel)) {
								callback.channelUpdated(channel);
							}
						} catch (IOException e) {
							Log.d("RestReader", "Error on reading channel. Reason: " + e.getMessage());
						}
					}
				}
				handler.sendEmptyMessage(0);
				if (callback != null) {
					callback.cycleFinished();
				}
			} catch (InterruptedException e) {
			}
		}
		runningThread = null;
	}

	/**
	 * Starts the runner. Does nothing if the runner is still alive
	 * 
	 * @param handler
	 *            Handler for notifying a ListView of updates
	 * @param client
	 *            rest client used to perform read requests on the server
	 * @param interval
	 *            time between the start of two read cycles in milliseconds
	 */
	public void start(Handler handler, RestClient client, int interval) {
		rest = client;
		this.interval = interval > 0 ? interval : 1;
		this.handler = handler;
		new Thread() {
			public void run() {
				if (runningThread != null && !running) {
					try {
						runningThread.join();
					} catch (InterruptedException e) {
					}
				}

				running = true;
				runningThread = new Thread(ChannelUpdater.this);
				runningThread.setName("ChannelUpdater");
				runningThread.setDaemon(true);
				runningThread.setPriority(Thread.NORM_PRIORITY - 1);
				runningThread.start();
			};

		}.start();
	}

	/**
	 * Calls the runner to stop at the next opportunity
	 */
	public void stop() {
		if (runningThread != null) {
			running = false;
			runningThread.interrupt();
		}
	}

	/**
	 * Bypass the waiting interval and directly start a single read cycle
	 */
	public void pollUpdate() {
		if (running) {
			runningThread.interrupt();
		}
	}

	/**
	 * Interface used by {@link ChannelUpdater} to notify about updated channels
	 * 
	 * @author Karsten Müller-Bier
	 */
	public interface UpdaterCallback {
		/**
		 * Notifies that a channel has been updated with a new record
		 * 
		 * @param c
		 *            updated channel
		 */
		void channelUpdated(Channel c);

		/**
		 * Notifies that a read cycle has ended. Time consuming tasks should be started here rather than inside {@link
		 * channelUpdated(Channel)}
		 */
		void cycleFinished();

		/**
		 * @return Collection of all channels the callback listens on. Only changes of channels that are in this
		 *         collection are notified by {@link channelUpdated(Channel)}
		 */
		Collection<Channel> getListenedChannels();
	}
}
