/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.openmuc.framework.driver.dummy;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;

import org.openmuc.framework.config.ArgumentSyntaxException;
import org.openmuc.framework.config.ChannelScanInformation;
import org.openmuc.framework.config.DeviceScanInformation;
import org.openmuc.framework.data.DoubleValue;
import org.openmuc.framework.data.Flag;
import org.openmuc.framework.data.Record;
import org.openmuc.framework.data.ValueType;
import org.openmuc.framework.driver.spi.ChannelRecordContainer;
import org.openmuc.framework.driver.spi.ChannelValueContainer;
import org.openmuc.framework.driver.spi.ConnectionException;
import org.openmuc.framework.driver.spi.DeviceConnection;
import org.openmuc.framework.driver.spi.DriverService;
import org.openmuc.framework.driver.spi.RecordsReceivedListener;
import org.osgi.service.component.ComponentContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public final class DummyDriver implements DriverService {

	private final static Logger logger = LoggerFactory.getLogger(DummyDriver.class);

	private final HashMap<Object, DummyListener> listenersByAddress = new HashMap<Object, DummyListener>();

	// Double p = Math.random();

	boolean applianceStateON = false;
	long applianceSwitchedOnTime;
	long applianceSwitchedOffTime;
	long applianceActiveTime;
	long applianceInactiveTime;
	boolean initApplianceState = true;

	protected void activate(ComponentContext context) {
		logger.info("Activating Dummy Driver");
	}

	protected void deactivate(ComponentContext context) {
		logger.info("Deactivating Dummy Driver");
		for (DummyListener listener : listenersByAddress.values()) {
			listener.shutdown();
		}
	}

	@Override
	public String getId() {
		return "dummy";
	}

	@Override
	public String getDescription() {
		return "This is just a dummy driver that returns dummy values and acts as a sink for written values";
	}

	@Override
	public List<DeviceScanInformation> scanForDevices(String interfaceAddress, String scanFilter, String settings,
			int timeout) throws UnsupportedOperationException, ArgumentSyntaxException, IOException {
		List<DeviceScanInformation> informationList = new ArrayList<DeviceScanInformation>(2);
		informationList.add(new DeviceScanInformation("dummy/device/address/1", "dummy-device1",
				"This is a dummy device."));
		informationList.add(new DeviceScanInformation("dummy/device/address/2", "dummy-device2",
				"This is a dummy device."));
		return informationList;
	}

	@Override
	public List<ChannelScanInformation> scanForChannels(DeviceConnection connection, int timeout)
			throws UnsupportedOperationException, ConnectionException {
		List<ChannelScanInformation> informationList = new ArrayList<ChannelScanInformation>(2);
		informationList.add(new ChannelScanInformation("dummy/channel/address/1", "", ValueType.DOUBLE, null));
		informationList.add(new ChannelScanInformation("dummy/channel/address/2", "", ValueType.BYTE_STRING, 5));
		return informationList;
	}

	@Override
	public Object connect(String interfaceAddress, String deviceAddress, String settings, int timeout)
			throws ConnectionException {
		logger.info("Connecting to interface:" + interfaceAddress + ", device:" + deviceAddress);

		// create you connection object:
		// Object newConnection = new Object();

		return interfaceAddress + "#" + deviceAddress;
	}

	@Override
	public void disconnect(DeviceConnection connection) {
		logger.info("Disconnecting from interface#device:" + (String) connection.getConnectionHandle());
	}

	@Override
	public Object read(DeviceConnection connection, List<ChannelRecordContainer> containers,
			Object containerListHandle, String samplingGroup, int timeout) throws UnsupportedOperationException {

		long receiveTime = System.currentTimeMillis();
		double value;
		for (ChannelRecordContainer channel : containers) {

			if (channel.getChannelAddress().equals("dummy/channel/address/voltage")) {
				// // voltage: 227 V - 233 V
				// value = Math.random() * (233 - 227) + 227;
				// voltage: 10 V - 12 V
				value = Math.random() * (12 - 10) + 10;
			}
			else if (channel.getChannelAddress().equals("dummy/channel/address/current")) {
				// current: 0.3 A - 0,7 A
				value = Math.random() * (0.7 - 0.3) + 0.3;
				setApplianceState();
				if (applianceStateON) {
					// add 5 A if appliance is on
					value += 1.5;
				}
			}
			else {
				// a random value between -10 and +10 for any other channel address
				value = Math.random() * (10 - (-10)) + (-10);
			}
			channel.setRecord(new Record(new DoubleValue(value), receiveTime));
		}
		// p += 1.0 / 90 % 2 * Math.PI;
		return null;
	}

	@Override
	public void startListening(DeviceConnection connection, List<ChannelRecordContainer> containers,
			RecordsReceivedListener listener) throws UnsupportedOperationException {
		DummyListener listenerThread = listenersByAddress.get(connection);
		if (containers.size() == 0) {
			if (listenerThread != null) {
				listenerThread.shutdown();
			}
		}
		else {
			if (listenerThread == null) {
				listenerThread = new DummyListener(containers, listener);
				listenerThread.start();
				listenersByAddress.put(connection, listenerThread);
			}
			else {
				listenerThread.setNewContainers(containers);
			}
		}
	}

	@Override
	public Object write(DeviceConnection connection, List<ChannelValueContainer> containers,
			Object containerListHandle, int timeout) throws UnsupportedOperationException, ConnectionException {
		try {
			Thread.sleep(500);
		} catch (InterruptedException e) {
		}
		for (ChannelValueContainer valueContainer : containers) {
			valueContainer.setFlag(Flag.VALID);
		}
		return null;
	}

	private void setApplianceState() {

		long currentTime = new Date().getTime();

		if (!applianceStateON) {
			// if appliance then leave it off for a certain time and then turn it on
			if ((applianceSwitchedOffTime + applianceInactiveTime) < currentTime || initApplianceState) {
				initApplianceState = false;
				applianceActiveTime = (long) (Math.random() * (10000 - 1000) + 1000);
				applianceSwitchedOnTime = new Date().getTime();
				applianceStateON = true;
			}
		}
		else {
			// if appliance on then turn it off after a certain time
			if ((applianceSwitchedOnTime + applianceActiveTime) < currentTime) {
				applianceInactiveTime = (long) (Math.random() * (10000 - 1000) + 1000);
				applianceSwitchedOffTime = new Date().getTime();
				applianceStateON = false;
			}
		}
	}
}
