/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.framework.demo.restclient;

import java.io.IOException;
import java.util.List;

import org.openmuc.framework.demo.restclient.data.Channel;
import org.openmuc.framework.demo.restclient.data.Record;
import org.restlet.resource.ResourceException;

/**
 * Basic application starter class that parses the command line parameter and branches execution depending on the parsed
 * parameters
 * 
 * @author Karsten Müller-Bier
 */
public class ClientStart {

	public static void main(String[] args) throws InterruptedException {
		if (args.length < 2) {
			String prgName = System.getProperty("sun.java.command");
			System.out.println("Usage: ");
			System.out.println("\"" + prgName + " <server> -list\"\n        --- list all available channels");
			System.out.println("\"" + prgName + " <server> <channel>\"\n        --- continously read from <channel>");
			System.out
					.println("\"" + prgName + " <server> <channel> <value>\"\n        --- write <value> to <channel>");
			System.out.println("\"" + prgName
					+ " <server> <channel> history <start> <end>\"\n        --- write <value> to <channel>");
			System.out.println();
			System.out.println("<server>: Name and port of the server");
			System.out.println("<channel>: Label of the channel to read/write");
			System.out.println("<value>: Numerical value that shall be written");
			System.out.println("<start>/<end>: Milliseconds since start of epoch");
			return;
		}

		String server = args[0];
		boolean listChannels = false;
		boolean useHttps = false;
		boolean history = false;
		String channel = "";
		Double newValue = null;
		long start = 0, end = 0;

		for (int i = 1; i < args.length; i++) {
			if (args[i].toLowerCase().equals("-list")) {
				listChannels = true;
			}
			else if (channel.isEmpty()) {
				channel = args[i];
			}
			else if ("history".equalsIgnoreCase(args[i])) {
				history = true;
				start = Long.parseLong(args[i + 1]);
				if (args.length > i + 2) {
					end = Long.parseLong(args[i + 2]);
				}
				else {
					end = System.currentTimeMillis();
				}

				break;
			}
			else {
				try {
					newValue = Double.parseDouble(args[i]);
					useHttps = true;
				} catch (NumberFormatException e) {
					System.out.println("Only numbers are allowed as new value. Aborting");
					return;
				}
			}
		}

		RestClient client = new RestClient(server, useHttps);

		try {
			if (listChannels) {
				List<Channel> channels = client.listChannels();
				System.out.println("Available channels...\n");
				for (Channel c : channels) {
					printChannel(c);
				}
				System.out.println("\n");
			}

			if (!channel.isEmpty()) {
				Channel targetChannel = new Channel();
				targetChannel.setLabel(channel);
				if (history) {
					System.out.println("History of " + channel + ":");
					List<Record> records = client.readChannelHistory(targetChannel, start, end);
					for (Record r : records) {
						printRecord(r);
					}
				}
				else if (newValue == null) {
					System.out.println("Begin continously reading " + channel + ". Press CTRL+C to quit\n");
					while (true) {
						if (client.readChannel(targetChannel)) {
							printChannel(targetChannel);
						}
						Thread.sleep(1000);
					}
				}
				else {
					System.out.println("Writing new value to " + channel);
					client.writeChannel(targetChannel, newValue);
					System.out.print("Channel written. Result:\n\t");
					printChannel(targetChannel);
				}
			}
		} catch (ResourceException ex) {
			System.out.println("\nError on GET/PUT. Server sent " + ex.getStatus().getCode() + " "
					+ ex.getStatus().getReasonPhrase());
		} catch (IOException e) {
			System.out.println("\nError on parsing JSON. Reason: " + e.getMessage());
		}
	}

	private static void printChannel(Channel channel) {
		Record record = channel.getRecord();

		StringBuilder sb = new StringBuilder();
		sb.append("Label: ").append(channel.getLabel());
		sb.append(", Timestamp: ").append(record.getTimestamp());
		sb.append(", Flag: ").append(record.getFlag());
		sb.append(", Value: ").append(record.getValue());
		System.out.println(sb.toString());
	}

	private static void printRecord(Record record) {
		StringBuilder sb = new StringBuilder();
		sb.append("Timestamp: ").append(record.getTimestamp());
		sb.append(", Flag: ").append(record.getFlag());
		sb.append(", Value: ").append(record.getValue());
		System.out.println(sb.toString());
	}
}
