/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.framework.demo.restclient;

import java.io.IOException;
import java.util.List;
import java.util.logging.Level;

import org.openmuc.framework.demo.restclient.data.Channel;
import org.openmuc.framework.demo.restclient.data.Record;
import org.openmuc.framework.demo.restclient.helper.InsecureSslContextFactory;
import org.restlet.Context;
import org.restlet.data.ChallengeScheme;
import org.restlet.data.MediaType;
import org.restlet.representation.Representation;
import org.restlet.representation.StringRepresentation;
import org.restlet.resource.ClientResource;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;

/**
 * Client-side endpoint of a RESTful communication.
 * 
 * A RestClient manages all objects needed to communicate with the RESTful Webservice defined in the
 * openmuc-webservice-rest project
 * 
 * @author Karsten Müller-Bier
 */
public class RestClient {

	private final ObjectMapper mapper;
	private final ClientResource client;

	/**
	 * @param server
	 *            The URL of the server
	 * @param useHttps
	 */
	public RestClient(String server, boolean useHttps) {
		Context ctx = new Context();
		if (useHttps) {
			// Create special context for the ClientResource to allow self signed certificates
			// Beware that this is a ugly hack and creates security risks. Do NOT use this in production code. There is
			// a reason that self signed certificates aren't trusted at default...
			ctx.getAttributes().put("sslContextFactory", new InsecureSslContextFactory());
		}
		server = fixServerString(server, useHttps);

		// Initialize Jackson
		mapper = new ObjectMapper();

		// Initialize Restlet Client
		client = new ClientResource(ctx, server);
		client.addSegment("rest").addSegment("channel"); // URL is now "http(s)://<server>/rest/channel"
		client.getLogger().setLevel(Level.WARNING); // Disables notifications when HttpClients are created
	}

	/**
	 * Adds the right protocol prefix to the server URL, if it's missing
	 * 
	 * @param server
	 *            The URL of the server
	 * @param useHttps
	 * @return The fixed Address
	 */
	private String fixServerString(String server, boolean useHttps) {
		if (useHttps && !server.startsWith("https://")) {
			return "https://" + server;
		}
		else if (!useHttps && !server.startsWith("http://")) {
			return "http://" + server;
		}
		return server;
	}

	/**
	 * Retrieves a list of all registered channels on the server. The list is deserialized using a TypeReference object.
	 * These objects are needed if the JSON Stream has to be deserialized into a generic class.
	 * 
	 * @return The list of remote channels
	 * @throws IOException
	 */
	public List<Channel> listChannels() throws IOException {
		Representation rep = client.get();
		List<Channel> channels = mapper.readValue(rep.getStream(), new TypeReference<List<Channel>>() {
		});
		return channels;
	}

	/**
	 * Retrieves the latest record of a channel. The record is deserialized using the class as reference. For
	 * deserializing Generics, see {@link RestClient#listChannels()}.
	 * 
	 * To read the latest record, a copy of the ClientResource instance variable is copied first and the label of the
	 * channel to read is added to the copy. After calling GET and retrieving the record, the copied ClientResourced is
	 * cleaned up using {@link ClientResource#release()}.
	 * 
	 * @param channel
	 *            The channel to read
	 * @return true if the received sample is newer than the previous one
	 * @throws IOException
	 */
	public boolean readChannel(Channel channel) throws IOException {
		ClientResource channelClient = new ClientResource(client);
		channelClient.addSegment(channel.getLabel());

		Representation rep = channelClient.get();
		Record newRecord = mapper.readValue(rep.getStream(), Record.class);
		channelClient.release();
		if (channel.getRecord().getTimestamp() < newRecord.getTimestamp()) {
			channel.setRecord(newRecord);
			return true;
		}
		return false;
	}

	public List<Record> readChannelHistory(Channel channel, long start, long end) throws IOException {
		ClientResource channelClient = new ClientResource(client);
		channelClient.addSegment(channel.getLabel()).addSegment("history");
		channelClient.addQueryParameter("from", "" + start);
		channelClient.addQueryParameter("until", "" + end);

		Representation rep = channelClient.get();
		List<Record> result = mapper.readValue(rep.getStream(), new TypeReference<List<Record>>() {
		});
		channelClient.release();

		return result;
	}

	/**
	 * Writes a new value to the remote channel.
	 * 
	 * The procedure is similar to {@link RestClient#listChannels()}, except that an additional challenge response is
	 * added to the client. This challenge response contains the username and password of the client used for
	 * authentication on server side.
	 * 
	 * @see readChannel
	 * @param channel
	 *            The channel that shall be written
	 * @param newValue
	 *            The new value to write into the channel
	 * @throws IOException
	 */
	public void writeChannel(Channel channel, double newValue) throws IOException {
		ClientResource channelClient = new ClientResource(client);
		channelClient.addSegment(channel.getLabel());
		channelClient.setChallengeResponse(ChallengeScheme.HTTP_BASIC, "admin", "admin");

		Record record = new Record();
		record.setValue(newValue);

		String json = mapper.writeValueAsString(record);
		Representation rep = channelClient.put(new StringRepresentation(json, MediaType.APPLICATION_JSON));
		channel.setRecord(mapper.readValue(rep.getStream(), Record.class));
		channelClient.release();
	}
}
