/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.openmuc.framework.demo.simpleapp;

import java.util.Date;
import java.util.concurrent.CountDownLatch;

import org.openmuc.framework.data.DoubleValue;
import org.openmuc.framework.data.Record;
import org.openmuc.framework.dataaccess.Channel;
import org.openmuc.framework.dataaccess.DataAccessService;
import org.openmuc.framework.dataaccess.RecordListener;
import org.osgi.service.component.ComponentContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public final class SimpleDemoApp extends Thread implements RecordListener {

	private final static Logger logger = LoggerFactory.getLogger(SimpleDemoApp.class);

	private volatile CountDownLatch threadFinishedSignal;
	private volatile boolean deactivatedSignal;

	DataAccessService dataAccessService;
	Channel channel1;
	Channel channel2;
	Channel channel3;

	public static final String channelLabel1 = "Voltage Channel";
	public static final String channelLabel2 = "Current Channel";
	public static final String channelLabel3 = "Power Channel";

	int printCounter; // for slowing down the output of the console

	protected void activate(ComponentContext context) {
		logger.info("Activating Demo App");
		start();

	}

	protected void deactivate(ComponentContext context) {
		logger.info("Deactivating Demo App");
		threadFinishedSignal = new CountDownLatch(1);
		deactivatedSignal = true;
		interrupt();
		try {
			threadFinishedSignal.await();
		} catch (InterruptedException e) {
		}
	}

	protected void setDataAccessService(DataAccessService dataAccessService) {
		this.dataAccessService = dataAccessService;
	}

	protected void unsetDataAccessService(DataAccessService dataAccessService) {
		this.dataAccessService = null;
	}

	@Override
	public void run() {

		logger.info("Demo App started running...");

		if (deactivatedSignal) {
			logger.info("DemoApp thread interrupted: will stop");
			threadFinishedSignal.countDown();
			return;
		}

		channel1 = dataAccessService.getChannel(channelLabel1);
		channel2 = dataAccessService.getChannel(channelLabel2);
		channel3 = dataAccessService.getChannel(channelLabel3);

		while (channel1 == null || channel2 == null || channel3 == null) {
			String errorMessage = "Channel " + channelLabel1 + " or " + channelLabel2 + " or " + channelLabel3
					+ " not found, will try again in 5 seconds.";
			logger.error(errorMessage);

			try {
				Thread.sleep(5000);
			} catch (InterruptedException e) {
				logger.info("DemoApp thread interrupted: will stop");
				threadFinishedSignal.countDown();
				return;
			}

			channel1 = dataAccessService.getChannel(channelLabel1);
			channel2 = dataAccessService.getChannel(channelLabel2);
			channel3 = dataAccessService.getChannel(channelLabel3);
		}

		// channel1.addListener(this);

		// WriteValueContainer writeContainer = channel1.getWriteContainer();
		// writeContainer.setValue(new DoubleValue(23.0));
		//
		// List<WriteValueContainer> writeContainerList = new ArrayList<WriteValueContainer>(1);
		// writeContainerList.add(writeContainer);

		while (true) {

			if (deactivatedSignal) {
				logger.info("DemoApp thread interrupted: will stop");
				threadFinishedSignal.countDown();
				return;
			}

			// List<Record> records = null;
			// try {
			// records = channel1.getLoggedValues(System.currentTimeMillis() - 50000);
			// } catch (DataLoggerNotAvailableException e1) {
			// logger.error("Unable to get logged values: log source unavailable");
			// } catch (IOException e1) {
			// logger.error("Unable to get logged values: IOException: " + e1.getMessage());
			// }
			// if (records != null) {
			// for (Record record : records) {
			// logger.info("Got logged record: " + record);
			// }
			// }

			// Flag flag = channel1.write(new DoubleValue(33.0), 3000);
			// logger.info("Wrote to channel1, resulting flag was: " + flag);
			//
			// logger.info("Channel1 - latest sample: " + channel1.getLatestSample());
			// logger.info("Channel1 - channel state: " + channel1.getChannelState());
			// logger.info("Channel1 - device state: " + channel1.getDeviceState());
			//
			// logger.info("Channel2 - latest sample: " + channel2.getLatestSample());
			// logger.info("Channel2 - channel state: " + channel2.getChannelState());
			// logger.info("Channel2 - device state: " + channel2.getDeviceState());

			try {

				double valueChannel1 = channel1.getLatestSample().getValue().asDouble(); // voltage
				double valueChannel2 = channel2.getLatestSample().getValue().asDouble(); // current
				double valueChannel3 = valueChannel1 * valueChannel2; // power
				// channel3.write(new DoubleValue(valueChannel3), 1000);
				channel3.setLatestSample(new Record(new DoubleValue(valueChannel3), new Date().getTime()));
				String state;
				if (valueChannel2 > 1.5) {
					state = "ON";
				}
				else {
					state = "OFF";
				}

				printCounter++;
				if (printCounter > 20) {
					logger.debug(channelLabel1 + ": " + valueChannel1 + "  " + channelLabel2 + ": " + valueChannel2
							+ "  " + channelLabel3 + ": " + valueChannel3 + "  state: " + state);
					printCounter = 0;
				}

			} catch (Exception e) {
				logger.error("somthing went wrong...");
			}

			try {
				Thread.sleep(100);
			} catch (InterruptedException e) {
				logger.info("DemoApp thread interrupted: will stop");
				channel1.removeListener(this);
				threadFinishedSignal.countDown();
				return;
			}
		}

	}

	@Override
	public void newRecord(Channel channel, Record record) {
		logger.info("Record listener got new record for channel: " + channel.getLabel() + " : " + record);
	}
}
