/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.openmuc.framework.core.datamanager;

import org.openmuc.framework.config.ChannelConfig;
import org.openmuc.framework.config.DeviceConfig;
import org.openmuc.framework.config.NamingCollisionException;
import org.openmuc.framework.config.ParseException;
import org.openmuc.framework.data.ValueType;
import org.openmuc.framework.dataaccess.ChannelState;
import org.openmuc.framework.datalogger.spi.LogChannel;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public final class ChannelConfigImpl implements ChannelConfig, LogChannel {

	String id;
	String channelAddress;
	String label = null;
	String description = null;
	String unit = null;
	Boolean readable = null;
	Boolean writable = null;
	ValueType valueType = null;
	Integer valueTypeLength = null;
	Double scalingFactor = null;
	Double valueOffset = null;
	Boolean listening = null;
	Integer samplingInterval = null;
	Integer samplingTimeOffset = null;
	String samplingGroup = null;
	Integer loggingInterval = null;
	Integer loggingTimeOffset = null;
	boolean disabled = DISABLED_DEFAULT;

	ChannelImpl channel;
	DeviceConfigImpl deviceParent;

	ChannelState state;

	public ChannelConfigImpl(String channelAddress, String id, DeviceConfigImpl deviceParent) {
		this.channelAddress = channelAddress;
		this.id = id;
		this.deviceParent = deviceParent;
	}

	@Override
	public String getId() {
		return id;
	}

	@Override
	public String getLabel() {
		return label;
	}

	@Override
	public void setLabel(String label) throws NamingCollisionException {
		if (this.label == label) {
			return;
		}
		deviceParent.driverParent.rootConfigParent.addChannelLabel(label, this);
		this.label = label;

	}

	@Override
	public void forceLabel(String label) {
		if (this.label == label) {
			return;
		}
		ChannelConfigImpl oldChannelConfig = deviceParent.driverParent.rootConfigParent.channelConfigsByLabel.put(
				label, this);
		if (oldChannelConfig != null) {
			oldChannelConfig.label = null;
		}

	}

	@Override
	public String getChannelAddress() {
		return channelAddress;
	}

	@Override
	public void setChannelAddress(String address) throws NamingCollisionException {
		if (address == null) {
			throw new IllegalArgumentException("The channel address may not be null");
		}
		deviceParent.changeChannelAddress(channelAddress, address);
		channelAddress = address;
	}

	@Override
	public String getDescription() {
		return description;
	}

	@Override
	public void setDescription(String description) {
		this.description = description;
	}

	@Override
	public String getUnit() {
		return unit;
	}

	@Override
	public void setUnit(String unit) {
		this.unit = unit;
	}

	@Override
	public Boolean isReadable() {
		return readable;
	}

	@Override
	public void setReadable(Boolean readable) {
		this.readable = readable;
	}

	@Override
	public Boolean isWritable() {
		return writable;
	}

	@Override
	public void setWritable(Boolean writable) {
		this.writable = writable;
	}

	@Override
	public ValueType getValueType() {
		return valueType;
	}

	@Override
	public void setValueType(ValueType valueType) {
		this.valueType = valueType;
	}

	@Override
	public Integer getValueTypeLength() {
		return valueTypeLength;
	}

	@Override
	public void setValueTypeLength(Integer length) {
		valueTypeLength = length;
	}

	@Override
	public Double getScalingFactor() {
		return scalingFactor;
	}

	@Override
	public void setScalingFactor(Double factor) {
		if (factor == null || factor != 0) {
			scalingFactor = factor;
		}
	}

	@Override
	public Double getValueOffset() {
		return valueOffset;
	}

	@Override
	public void setValueOffset(Double offset) {
		valueOffset = offset;
	}

	@Override
	public Boolean isListening() {
		return listening;
	}

	@Override
	public void setListening(Boolean listening) {
		this.listening = listening;
	}

	@Override
	public Integer getSamplingInterval() {
		return samplingInterval;
	}

	@Override
	public void setSamplingInterval(Integer samplingInterval) {
		this.samplingInterval = samplingInterval;
	}

	@Override
	public Integer getSamplingTimeOffset() {
		return samplingTimeOffset;
	}

	@Override
	public void setSamplingTimeOffset(Integer samplingTimeOffset) {
		this.samplingTimeOffset = samplingTimeOffset;
	}

	@Override
	public String getSamplingGroup() {
		return samplingGroup;
	}

	@Override
	public void setSamplingGroup(String group) {
		samplingGroup = group;
	}

	@Override
	public Integer getLoggingInterval() {
		return loggingInterval;
	}

	@Override
	public void setLoggingInterval(Integer loggingInterval) {
		this.loggingInterval = loggingInterval;
	}

	@Override
	public Integer getLoggingTimeOffset() {
		return loggingTimeOffset;
	}

	@Override
	public void setLoggingTimeOffset(Integer loggingTimeOffset) {
		this.loggingTimeOffset = loggingTimeOffset;
	}

	@Override
	public Boolean isDisabled() {
		return disabled || deviceParent.isDisabled();
	}

	@Override
	public void setDisabled(Boolean disabled) {
		this.disabled = disabled;
	}

	@Override
	public void delete() {
		deviceParent.removeChannel(this);
		clear();
	}

	void clear() {
		deviceParent = null;
	}

	static void addChannelFromDomNode(Node channelNode, DeviceConfig parentDevice) throws ParseException {

		String childName;
		String channelAddress = null;
		String channelId = null;

		NodeList channelChildren = channelNode.getChildNodes();

		for (int l = 0; l < channelChildren.getLength(); l++) {
			Node childNode = channelChildren.item(l);
			childName = childNode.getNodeName();

			if (childName.equals("channelAddress")) {
				channelAddress = childNode.getTextContent();
			}
			else if (childName.equals("id")) {
				channelId = childNode.getTextContent();
			}
		}

		if (channelAddress == null) {
			throw new ParseException("Required \"channelAddress\" element not found.");
		}

		ChannelConfigImpl config;
		if (channelId == null) {
			try {
				config = (ChannelConfigImpl) parentDevice.addChannel(channelAddress);
			} catch (NamingCollisionException e) {
				throw new ParseException(e);
			}
		}
		else {
			try {
				config = (ChannelConfigImpl) parentDevice.addChannel(channelAddress, channelId);
			} catch (NamingCollisionException e) {
				throw new ParseException(e);
			}
		}

		for (int i = 0; i < channelChildren.getLength(); i++) {
			Node childNode = channelChildren.item(i);

			childName = childNode.getNodeName();

			if (childName.equals("#text") || childName.equals("channelAddress") || childName.equals("id")) {
				continue;
			}
			else if (childName.equals("label")) {
				try {
					config.setLabel(childNode.getTextContent());
				} catch (Exception e) {
					throw new ParseException(e);
				}
			}
			else if (childName.equals("description")) {
				config.setDescription(childNode.getTextContent());
			}
			else if (childName.equals("unit")) {
				config.setUnit(childNode.getTextContent());
			}
			else if (childName.equals("readable")) {
				String readableString = childNode.getTextContent().toLowerCase();
				if (readableString.equals("true")) {
					config.readable = true;
				}
				else if (readableString.equals("false")) {
					config.readable = false;
				}
			}
			else if (childName.equals("writable")) {
				String writableString = childNode.getTextContent().toLowerCase();
				if (writableString.equals("true")) {
					config.writable = true;
				}
				else if (writableString.equals("false")) {
					config.writable = false;
				}
			}
			else if (childName.equals("valueType")) {
				String valueTypeString = childNode.getTextContent().toUpperCase();
				if (valueTypeString.equals(ValueType.DOUBLE.toString())) {
					config.valueType = ValueType.DOUBLE;
				}
				else if (valueTypeString.equals(ValueType.BYTE_STRING.toString())) {
					config.valueType = ValueType.BYTE_STRING;
					String valueTypeLengthString = getAttributeValue(childNode, "length");
					if (valueTypeLengthString == null) {
						throw new ParseException("length of BYTE_STRING value type was not specified");
					}
					config.valueTypeLength = timeStringToMillis(valueTypeLengthString);
				}
				else {
					throw new ParseException("found unknown channel value type:" + valueTypeString);
				}
			}
			else if (childName.equals("scalingFactor")) {
				config.scalingFactor = Double.parseDouble(childNode.getTextContent());
			}
			else if (childName.equals("valueOffset")) {
				config.valueOffset = Double.parseDouble(childNode.getTextContent());
			}
			else if (childName.equals("listening")) {
				String listeningString = childNode.getTextContent().toLowerCase();
				if (listeningString.equals("true")) {
					config.listening = true;
				}
				else if (listeningString.equals("false")) {
					config.listening = false;
				}
			}
			else if (childName.equals("samplingInterval")) {
				config.samplingInterval = timeStringToMillis(childNode.getTextContent());
			}
			else if (childName.equals("samplingTimeOffset")) {
				config.samplingTimeOffset = timeStringToMillis(childNode.getTextContent());
			}
			else if (childName.equals("samplingGroup")) {
				config.setSamplingGroup(childNode.getTextContent());
			}
			else if (childName.equals("loggingInterval")) {
				config.loggingInterval = timeStringToMillis(childNode.getTextContent());
			}
			else if (childName.equals("loggingTimeOffset")) {
				config.loggingTimeOffset = timeStringToMillis(childNode.getTextContent());
			}
			else if (childName.equals("disabled")) {
				String disabledString = childNode.getTextContent().toLowerCase();
				if (disabledString.equals("true")) {
					config.disabled = true;
				}
				else if (disabledString.equals("false")) {
					config.disabled = false;
				}
			}

		}

	}

	Element getDomElement(Document document) {
		Element parentElement = document.createElement("channel");

		Element childElement;

		childElement = document.createElement("id");
		childElement.setTextContent(id);
		parentElement.appendChild(childElement);

		childElement = document.createElement("channelAddress");
		childElement.setTextContent(channelAddress);
		parentElement.appendChild(childElement);

		if (label != null) {
			childElement = document.createElement("label");
			childElement.setTextContent(label);
			parentElement.appendChild(childElement);
		}

		if (description != null) {
			childElement = document.createElement("description");
			childElement.setTextContent(description);
			parentElement.appendChild(childElement);
		}

		if (unit != null) {
			childElement = document.createElement("unit");
			childElement.setTextContent(unit);
			parentElement.appendChild(childElement);
		}

		if (readable != null) {
			childElement = document.createElement("readable");
			if (readable) {
				childElement.setTextContent("true");
			}
			else {
				childElement.setTextContent("false");
			}
			parentElement.appendChild(childElement);
		}

		if (writable != null) {
			childElement = document.createElement("writable");
			if (writable) {
				childElement.setTextContent("true");
			}
			else {
				childElement.setTextContent("false");
			}
			parentElement.appendChild(childElement);
		}

		if (valueType != null) {
			childElement = document.createElement("valueType");
			childElement.setTextContent(valueType.toString());
			if (valueType == ValueType.BYTE_STRING) {
				childElement.setAttribute("length", valueTypeLength.toString());
			}
			parentElement.appendChild(childElement);
		}

		if (scalingFactor != null) {
			childElement = document.createElement("scalingFactor");
			childElement.setTextContent(Double.toString(scalingFactor));
			parentElement.appendChild(childElement);
		}

		if (valueOffset != null) {
			childElement = document.createElement("valueOffset");
			childElement.setTextContent(Double.toString(valueOffset));
			parentElement.appendChild(childElement);
		}

		if (listening != null) {
			childElement = document.createElement("listening");
			if (listening) {
				childElement.setTextContent("true");
			}
			else {
				childElement.setTextContent("false");
			}
			parentElement.appendChild(childElement);
		}

		if (samplingInterval != null) {
			childElement = document.createElement("samplingInterval");
			childElement.setTextContent(millisToTimeString(samplingInterval));
			parentElement.appendChild(childElement);
		}

		if (samplingTimeOffset != null) {
			childElement = document.createElement("samplingTimeOffset");
			childElement.setTextContent(millisToTimeString(samplingTimeOffset));
			parentElement.appendChild(childElement);
		}

		if (samplingGroup != null) {
			childElement = document.createElement("samplingGroup");
			childElement.setTextContent(samplingGroup);
			parentElement.appendChild(childElement);
		}

		if (loggingInterval != null) {
			childElement = document.createElement("loggingInterval");
			childElement.setTextContent(millisToTimeString(loggingInterval));
			parentElement.appendChild(childElement);
		}

		if (loggingTimeOffset != null) {
			childElement = document.createElement("loggingTimeOffset");
			childElement.setTextContent(millisToTimeString(loggingTimeOffset));
			parentElement.appendChild(childElement);
		}

		childElement = document.createElement("disabled");
		if (disabled) {
			childElement.setTextContent("true");
		}
		else {
			childElement.setTextContent("false");
		}
		parentElement.appendChild(childElement);

		return parentElement;
	}

	ChannelConfigImpl clone(DeviceConfigImpl clonedParentConfig) {
		ChannelConfigImpl configClone = new ChannelConfigImpl(channelAddress, id, clonedParentConfig);
		configClone.label = label;
		configClone.description = description;
		configClone.unit = unit;
		configClone.readable = readable;
		configClone.writable = writable;
		configClone.valueType = valueType;
		configClone.valueTypeLength = valueTypeLength;
		configClone.scalingFactor = scalingFactor;
		configClone.valueOffset = valueOffset;
		configClone.listening = listening;
		configClone.samplingInterval = samplingInterval;
		configClone.samplingTimeOffset = samplingTimeOffset;
		configClone.samplingGroup = samplingGroup;
		configClone.loggingInterval = loggingInterval;
		configClone.loggingTimeOffset = loggingTimeOffset;
		configClone.disabled = disabled;

		return configClone;
	}

	ChannelConfigImpl cloneWithDefaults(DeviceConfigImpl clonedParentConfig) {
		ChannelConfigImpl configClone = new ChannelConfigImpl(channelAddress, id, clonedParentConfig);

		configClone.label = label;

		if (description == null) {
			configClone.description = ChannelConfig.DESCRIPTION_DEFAULT;
		}
		else {
			configClone.description = description;
		}

		if (unit == null) {
			configClone.unit = ChannelConfig.UNIT_DEFAULT;
		}
		else {
			configClone.unit = unit;
		}

		if (readable == null) {
			configClone.readable = ChannelConfig.READABLE_DEFAULT;
		}
		else {
			configClone.readable = readable;
		}

		if (writable == null) {
			configClone.writable = ChannelConfig.WRITABLE_DEFAULT;
		}
		else {
			configClone.writable = writable;
		}

		if (valueType == null) {
			configClone.valueType = ChannelConfig.VALUE_TYPE_DEFAULT;
		}
		else {
			configClone.valueType = valueType;
		}

		if (valueTypeLength == null) {
			if (valueType == ValueType.DOUBLE) {
				configClone.valueTypeLength = 8;
			}
			else if (valueType == ValueType.BYTE_STRING) {
				configClone.valueTypeLength = ChannelConfig.BYTE_STRING_SIZE_DEFAULT;
			}
			else if (valueType == ValueType.BYTE) {
				configClone.valueTypeLength = 1;
			}
			else if (valueType == ValueType.BYTE) {
				configClone.valueTypeLength = 1;
			}
			else if (valueType == ValueType.FLOAT) {
				configClone.valueTypeLength = 4;
			}
			else if (valueType == ValueType.SHORT) {
				configClone.valueTypeLength = 2;
			}
			else if (valueType == ValueType.INTEGER) {
				configClone.valueTypeLength = 4;
			}
			else if (valueType == ValueType.LONG) {
				configClone.valueTypeLength = 8;
			}
			else if (valueType == ValueType.BOOLEAN) {
				configClone.valueTypeLength = 1;
			}
		}
		else {
			configClone.valueTypeLength = valueTypeLength;
		}

		configClone.scalingFactor = scalingFactor;
		configClone.valueOffset = valueOffset;

		if (listening == null) {
			configClone.listening = ChannelConfig.LISTENING_DEFAULT;
		}
		else {
			configClone.listening = listening;
		}

		if (samplingInterval == null) {
			configClone.samplingInterval = ChannelConfig.SAMPLING_INTERVAL_DEFAULT;
		}
		else {
			configClone.samplingInterval = samplingInterval;
		}

		if (samplingTimeOffset == null) {
			configClone.samplingTimeOffset = ChannelConfig.SAMPLING_TIME_OFFSET_DEFAULT;
		}
		else {
			configClone.samplingTimeOffset = samplingTimeOffset;
		}

		if (samplingGroup == null) {
			configClone.samplingGroup = ChannelConfig.SAMPLING_GROUP_DEFAULT;
		}
		else {
			configClone.samplingGroup = samplingGroup;
		}

		if (loggingInterval == null) {
			configClone.loggingInterval = ChannelConfig.LOGGING_INTERVAL_DEFAULT;
		}
		else {
			configClone.loggingInterval = loggingInterval;
		}

		if (loggingTimeOffset == null) {
			configClone.loggingTimeOffset = ChannelConfig.LOGGING_TIME_OFFSET_DEFAULT;
		}
		else {
			configClone.loggingTimeOffset = loggingTimeOffset;
		}

		configClone.disabled = disabled;

		return configClone;
	}

	static String getAttributeValue(Node element, String attributeName) {
		NamedNodeMap attributes = element.getAttributes();

		Node nameAttribute = attributes.getNamedItem(attributeName);

		if (nameAttribute == null) {
			return null;
		}
		return nameAttribute.getTextContent();
	}

	static String millisToTimeString(Integer time) {
		if (time == 0) {
			return "0";
		}
		if ((time % 1000) == 0) {
			time /= 1000;
			if ((time % 60) == 0) {
				time /= 60;
				if ((time % 60) == 0) {
					return time.toString().concat("h");
				}
				return time.toString().concat("m");
			}
			return time.toString().concat("s");
		}
		return time.toString().concat("ms");
	}

	static int timeStringToMillis(String timeString) throws ParseException {
		try {

			char lastChar = timeString.charAt(timeString.length() - 1);

			if (Character.isDigit(lastChar)) {
				return Integer.parseInt(timeString);
			}

			switch (lastChar) {
			case 's':
				if (timeString.charAt(timeString.length() - 2) == 'm') {
					return Integer.parseInt(timeString.substring(0, timeString.length() - 2));
				}
				return Integer.parseInt(timeString.substring(0, timeString.length() - 1)) * 1000;
			case 'm':
				return Integer.parseInt(timeString.substring(0, timeString.length() - 1)) * 60000;
			case 'h':
				return Integer.parseInt(timeString.substring(0, timeString.length() - 1)) * 3600000;
			default:
				throw new ParseException("unknown time string: " + timeString);
			}

		} catch (NumberFormatException e) {
			throw new ParseException(e);
		}

	}

	public boolean isSampling() {
		return samplingInterval != null && samplingInterval > 0;
	}

}
