/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.openmuc.framework.core.datamanager;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;

import org.openmuc.framework.config.ChannelConfig;
import org.openmuc.framework.config.DeviceConfig;
import org.openmuc.framework.config.DriverConfig;
import org.openmuc.framework.config.NamingCollisionException;
import org.openmuc.framework.config.ParseException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public final class DeviceConfigImpl implements DeviceConfig {

	String ifAndDevAddress;
	String deviceAddress;
	String interfaceAddress = null;
	String name = null;
	String description = null;
	String settings = null;

	Integer readTimeout = null;
	Integer connectTimeout = null;
	Integer connectRetryInterval = null;
	boolean disabled = DISABLED_DEFAULT;

	Device device = null;

	final HashMap<String, ChannelConfigImpl> channelConfigsByAddress;

	DriverConfigImpl driverParent;

	DeviceConfigImpl(String interfaceAddress, String deviceAddress, String address, DriverConfigImpl driverParent) {
		this.deviceAddress = deviceAddress;
		this.interfaceAddress = interfaceAddress;
		this.driverParent = driverParent;
		ifAndDevAddress = this.interfaceAddress + "#" + this.deviceAddress;
		channelConfigsByAddress = new LinkedHashMap<String, ChannelConfigImpl>();
	}

	public String getIfAndDevAddress() {
		return ifAndDevAddress;
	}

	@Override
	public String getDeviceAddress() {
		return deviceAddress;
	}

	@Override
	public String getInterfaceAddress() {
		return interfaceAddress;
	}

	@Override
	public void setAddresses(String interfaceAddress, String deviceAddress) throws NamingCollisionException {
		if (deviceAddress == null) {
			throw new IllegalArgumentException("The device address may not be null");
		}
		String newAddress;
		if (interfaceAddress == null) {
			newAddress = deviceAddress;
		}
		else {
			newAddress = interfaceAddress + "#" + deviceAddress;
		}
		driverParent.changeIfAndDevAddress(ifAndDevAddress, newAddress);
		ifAndDevAddress = newAddress;
		this.deviceAddress = deviceAddress;
		this.interfaceAddress = interfaceAddress;
	}

	@Override
	public String getName() {
		return name;
	}

	@Override
	public void setName(String name) {
		this.name = name;
	}

	@Override
	public String getDescription() {
		return description;
	}

	@Override
	public void setDescription(String description) {
		this.description = description;
	}

	@Override
	public String getSettings() {
		return settings;
	}

	@Override
	public void setSettings(String settings) {
		this.settings = settings;
	}

	@Override
	public Integer getReadTimeout() {
		return readTimeout;
	}

	@Override
	public void setReadTimeout(Integer timeout) {
		readTimeout = timeout;
	}

	@Override
	public Integer getConnectTimeout() {
		return connectTimeout;
	}

	@Override
	public void setConnectTimeout(Integer timeout) {
		connectTimeout = timeout;
	}

	@Override
	public Integer getConnectRetryInterval() {
		return connectRetryInterval;
	}

	@Override
	public void setConnectRetryInterval(Integer interval) {
		connectRetryInterval = interval;
	}

	@Override
	public Boolean isDisabled() {
		return disabled || driverParent.isDisabled();
	}

	@Override
	public void setDisabled(Boolean disabled) {
		this.disabled = disabled;
	}

	@Override
	public ChannelConfig addChannel(String channelAddress) throws NamingCollisionException {

		if (channelAddress == null) {
			throw new IllegalArgumentException("The channelAddress string may not be null");
		}
		if (channelConfigsByAddress.containsKey(channelAddress)) {
			throw new NamingCollisionException("Collision with channel address: " + channelAddress
					+ " in device with address: " + deviceAddress);
		}

		while (true) {
			String id = RootConfigImpl.randomString.nextString();
			ChannelConfigImpl newChannel = new ChannelConfigImpl(channelAddress, id, this);
			try {
				driverParent.rootConfigParent.addChannelId(id, newChannel);
			} catch (NamingCollisionException e) {
				continue;
			}
			channelConfigsByAddress.put(channelAddress, newChannel);
			return newChannel;
		}
	}

	@Override
	public ChannelConfig addChannel(String channelAddress, String id) throws NamingCollisionException {
		if (channelAddress == null || id == null) {
			throw new IllegalArgumentException("The channelAddress or id strings may not be null");
		}
		if (channelConfigsByAddress.containsKey(channelAddress)) {
			throw new NamingCollisionException("Collision with channel address: " + channelAddress
					+ " in device with address: " + deviceAddress);
		}

		ChannelConfigImpl newChannel = new ChannelConfigImpl(channelAddress, id, this);
		driverParent.rootConfigParent.addChannelId(id, newChannel);
		channelConfigsByAddress.put(channelAddress, newChannel);
		return newChannel;
	}

	@Override
	public ChannelConfig getOrAddChannel(String channelAddress) {
		try {
			return addChannel(channelAddress);
		} catch (NamingCollisionException e) {
			return channelConfigsByAddress.get(channelAddress);
		}
	}

	@Override
	public ChannelConfig getChannel(String channelAddress) {
		return channelConfigsByAddress.get(channelAddress);
	}

	@SuppressWarnings("unchecked")
	@Override
	public Collection<ChannelConfig> getChannels() {
		return (Collection<ChannelConfig>) (Collection<?>) Collections.unmodifiableCollection(channelConfigsByAddress
				.values());
	}

	@Override
	public void delete() {
		driverParent.removeDevice(ifAndDevAddress);
		clear();
	}

	void clear() {
		for (ChannelConfigImpl channelConfig : channelConfigsByAddress.values()) {
			channelConfig.clear();
		}
		driverParent = null;
	}

	static void addDeviceFromDomNode(Node driverChild, DriverConfig parentDriver) throws ParseException {

		NodeList deviceChildren = driverChild.getChildNodes();
		String deviceAddress = null;
		String interfaceAddress = null;

		for (int i = 0; i < deviceChildren.getLength(); i++) {
			Node deviceChild = deviceChildren.item(i);
			String childName = deviceChild.getNodeName();

			if (childName.equals("deviceAddress")) {
				deviceAddress = deviceChild.getTextContent();
			}
			else if (childName.equals("interfaceAddress")) {
				interfaceAddress = deviceChild.getTextContent();
			}

		}

		if (deviceAddress == null) {
			throw new ParseException("the deviceAddress was not defined");
		}

		DeviceConfigImpl config;
		try {
			config = (DeviceConfigImpl) parentDriver.addDevice(interfaceAddress, deviceAddress);
		} catch (Exception e) {
			throw new ParseException(e);
		}

		for (int i = 0; i < deviceChildren.getLength(); i++) {
			Node childNode = deviceChildren.item(i);

			String childName = childNode.getNodeName();

			if (childName.equals("#text") || childName.equals("deviceAddress") || childName.equals("interfaceAddress")) {
				continue;
			}
			else if (childName.equals("channel")) {
				ChannelConfigImpl.addChannelFromDomNode(childNode, config);
			}
			else if (childName.equals("name")) {
				try {
					config.setName(childNode.getTextContent());
				} catch (Exception e) {
					throw new ParseException(e);
				}
			}
			else if (childName.equals("description")) {
				config.setDescription(childNode.getTextContent());
			}
			else if (childName.equals("settings")) {
				config.setSettings(childNode.getTextContent());
			}
			else if (childName.equals("readTimeout")) {
				config.setReadTimeout(ChannelConfigImpl.timeStringToMillis(childNode.getTextContent()));
			}
			else if (childName.equals("connectTimeout")) {
				config.setConnectTimeout(ChannelConfigImpl.timeStringToMillis(childNode.getTextContent()));
			}
			else if (childName.equals("connectRetryInterval")) {
				config.setConnectRetryInterval(ChannelConfigImpl.timeStringToMillis(childNode.getTextContent()));
			}
			else if (childName.equals("disabled")) {
				String disabledString = childNode.getTextContent().toLowerCase();
				if (disabledString.equals("true")) {
					config.disabled = true;
				}
				else if (disabledString.equals("false")) {
					config.disabled = false;
				}
			}
			else {
				throw new ParseException("found unknown tag:" + childName);
			}
		}

	}

	Element getDomElement(Document document) {
		Element parentElement = document.createElement("device");

		Element childElement;

		childElement = document.createElement("deviceAddress");
		childElement.setTextContent(deviceAddress);
		parentElement.appendChild(childElement);

		if (interfaceAddress != null) {
			childElement = document.createElement("interfaceAddress");
			childElement.setTextContent(interfaceAddress);
			parentElement.appendChild(childElement);
		}

		if (name != null) {
			childElement = document.createElement("name");
			childElement.setTextContent(name);
			parentElement.appendChild(childElement);
		}

		if (description != null) {
			childElement = document.createElement("description");
			childElement.setTextContent(description);
			parentElement.appendChild(childElement);
		}

		if (settings != null) {
			childElement = document.createElement("settings");
			childElement.setTextContent(settings);
			parentElement.appendChild(childElement);
		}

		if (readTimeout != null) {
			childElement = document.createElement("readTimeout");
			childElement.setTextContent(ChannelConfigImpl.millisToTimeString(readTimeout));
			parentElement.appendChild(childElement);
		}

		if (connectTimeout != null) {
			childElement = document.createElement("connectTimeout");
			childElement.setTextContent(ChannelConfigImpl.millisToTimeString(connectTimeout));
			parentElement.appendChild(childElement);
		}

		if (connectRetryInterval != null) {
			childElement = document.createElement("connectRetryInterval");
			childElement.setTextContent(ChannelConfigImpl.millisToTimeString(connectRetryInterval));
			parentElement.appendChild(childElement);
		}

		childElement = document.createElement("disabled");
		if (disabled) {
			childElement.setTextContent("true");
		}
		else {
			childElement.setTextContent("false");
		}
		parentElement.appendChild(childElement);

		for (ChannelConfigImpl channelConfig : channelConfigsByAddress.values()) {
			parentElement.appendChild(channelConfig.getDomElement(document));
		}

		return parentElement;
	}

	void changeChannelAddress(String oldAddress, String newAddress) throws NamingCollisionException {
		if (channelConfigsByAddress.containsKey(newAddress)) {
			throw new NamingCollisionException("Collision with channel address: " + newAddress);
		}
		channelConfigsByAddress.put(newAddress, channelConfigsByAddress.remove(oldAddress));
	}

	void removeChannel(ChannelConfigImpl channelConfig) {
		channelConfigsByAddress.remove(channelConfig.channelAddress);
		driverParent.rootConfigParent.removeChannel(channelConfig);
	}

	DeviceConfigImpl clone(DriverConfigImpl clonedParentConfig) {
		DeviceConfigImpl configClone = new DeviceConfigImpl(interfaceAddress, deviceAddress, ifAndDevAddress,
				clonedParentConfig);
		configClone.name = name;
		configClone.description = description;
		configClone.settings = settings;
		configClone.readTimeout = readTimeout;
		configClone.connectTimeout = connectTimeout;
		configClone.connectRetryInterval = connectRetryInterval;
		configClone.disabled = disabled;
		for (ChannelConfigImpl channelConfig : channelConfigsByAddress.values()) {
			configClone.channelConfigsByAddress.put(channelConfig.channelAddress, channelConfig.clone(configClone));
		}
		return configClone;
	}

	DeviceConfigImpl cloneWithDefaults(DriverConfigImpl clonedParentConfig) {
		DeviceConfigImpl configClone = new DeviceConfigImpl(interfaceAddress, deviceAddress, ifAndDevAddress,
				clonedParentConfig);

		if (name == null) {
			configClone.name = DeviceConfig.NAME_DEFAULT;
		}
		else {
			configClone.name = name;
		}

		if (description == null) {
			configClone.description = DeviceConfig.DESCRIPTION_DEFAULT;
		}
		else {
			configClone.description = description;
		}

		if (settings == null) {
			configClone.settings = DeviceConfig.SETTINGS_DEFAULT;
		}
		else {
			configClone.settings = settings;
		}

		if (readTimeout == null) {
			configClone.readTimeout = clonedParentConfig.readTimeout;
		}
		else {
			configClone.readTimeout = readTimeout;
		}

		if (connectTimeout == null) {
			configClone.connectTimeout = clonedParentConfig.connectTimeout;
		}
		else {
			configClone.connectTimeout = connectTimeout;
		}

		if (connectRetryInterval == null) {
			configClone.connectRetryInterval = clonedParentConfig.connectRetryInterval;
		}
		else {
			configClone.connectRetryInterval = connectRetryInterval;
		}

		configClone.disabled = disabled;

		for (ChannelConfigImpl channelConfig : channelConfigsByAddress.values()) {
			configClone.channelConfigsByAddress.put(channelConfig.channelAddress,
					channelConfig.cloneWithDefaults(configClone));
		}
		return configClone;
	}
}
