/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.openmuc.framework.core.datamanager;

import java.util.List;

import org.openmuc.framework.data.Flag;
import org.openmuc.framework.data.Record;
import org.openmuc.framework.dataaccess.DeviceState;
import org.openmuc.framework.driver.spi.ChannelRecordContainer;
import org.openmuc.framework.driver.spi.ConnectionException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public abstract class DeviceReadTask extends DeviceTask {

	private final static Logger logger = LoggerFactory.getLogger(DeviceReadTask.class);

	List<ChannelRecordContainerImpl> selectedChannels;
	protected final int readTimeout;
	protected boolean methodNotExceptedExceptionThrown = false;
	protected boolean unknownDriverExceptionThrown = false;
	protected volatile boolean disabled = false;

	boolean running = false;
	boolean startedLate = false;

	protected DeviceReadTask(DataManager dataManager, Device device, List<ChannelRecordContainerImpl> selectedChannels,
			int readTimeout) {
		this.dataManager = dataManager;
		this.device = device;
		this.selectedChannels = selectedChannels;
		this.readTimeout = readTimeout;
	}

	@Override
	public final void run() {

		try {
			executeRead();
		} catch (UnsupportedOperationException e) {
			methodNotExceptedExceptionThrown = true;
		} catch (ConnectionException e) {
			// Connection to device lost. Signal to device instance and end task without notifying DataManager
			logger.warn("Connection to device {} lost because {}. Trying to reconnect...",
					device.deviceConfig.getIfAndDevAddress(), e.getMessage());
			device.disconnectedSignal();
			return;
		} catch (Exception e) {
			logger.warn("unexpected exception thrown by read funtion of driver ", e);
			unknownDriverExceptionThrown = true;
		}

		taskFinished();
	}

	public final void timeout() {
		if (!disabled) {
			disabled = true;
			if (startedLate) {
				for (ChannelRecordContainerImpl driverChannel : selectedChannels) {
					driverChannel.channel.setFlag(Flag.STARTED_LATE_AND_TIMED_OUT);
				}
			}
			else if (running) {
				for (ChannelRecordContainerImpl driverChannel : selectedChannels) {
					driverChannel.channel.setFlag(Flag.TIMEOUT);
				}
			}
			else {
				for (ChannelRecordContainerImpl driverChannel : selectedChannels) {
					driverChannel.channel.setFlag(Flag.DEVICE_OR_INTERFACE_BUSY);
				}
				device.removeTask(this);
			}

		}
	}

	@Override
	public final DeviceTaskType getType() {
		return DeviceTaskType.READ;
	}

	@Override
	public final void setDeviceState() {
		device.state = DeviceState.READING;
	}

	public final void deviceNotConnected() {
		for (ChannelRecordContainer recordContainer : selectedChannels) {
			recordContainer.setRecord(new Record(null, System.currentTimeMillis(), Flag.COMM_DEVICE_NOT_CONNECTED));
		}
		taskAborted();
	}

	protected abstract void executeRead() throws UnsupportedOperationException, ConnectionException;

	protected abstract void taskFinished();

	protected abstract void taskAborted();
}
