/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.openmuc.framework.core.datamanager;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map.Entry;

import org.openmuc.framework.config.DeviceConfig;
import org.openmuc.framework.config.DriverConfig;
import org.openmuc.framework.config.NamingCollisionException;
import org.openmuc.framework.config.ParseException;
import org.openmuc.framework.config.RootConfig;
import org.openmuc.framework.driver.spi.DriverService;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public final class DriverConfigImpl implements DriverConfig {

	String id;
	RootConfigImpl rootConfigParent;

	Integer readTimeout = null;
	Integer connectTimeout = null;
	Integer connectRetryInterval = null;
	boolean disabled = DISABLED_DEFAULT;

	final HashMap<String, DeviceConfigImpl> deviceConfigsByAddress = new LinkedHashMap<String, DeviceConfigImpl>();
	final HashMap<String, List<DeviceTask>> deviceTasksByInterfaceAddress = new LinkedHashMap<String, List<DeviceTask>>();

	DriverService activeDriver = null;

	DriverConfigImpl(String id, RootConfigImpl rootConfigParent) {
		this.id = id;
		this.rootConfigParent = rootConfigParent;
	}

	@Override
	public String getId() {
		return id;
	}

	@Override
	public void setId(String id) throws NamingCollisionException {
		if (id == null) {
			throw new IllegalArgumentException("The driver ID may not be null");
		}
		if (this.id == id) {
			return;
		}
		rootConfigParent.changeId(this.id, id);
		this.id = id;
	}

	@Override
	public Integer getReadTimeout() {
		return readTimeout;
	}

	@Override
	public void setReadTimeout(Integer timeout) {
		readTimeout = timeout;
	}

	@Override
	public Integer getConnectTimeout() {
		return connectTimeout;
	}

	@Override
	public void setConnectTimeout(Integer timeout) {
		connectTimeout = timeout;
	}

	@Override
	public Integer getConnectRetryInterval() {
		return connectRetryInterval;
	}

	@Override
	public void setConnectRetryInterval(Integer interval) {
		connectRetryInterval = interval;
	}

	@Override
	public Boolean isDisabled() {
		return disabled;
	}

	@Override
	public void setDisabled(Boolean disabled) {
		this.disabled = disabled;
	}

	@Override
	public DeviceConfig addDevice(String interfaceAddress, String deviceAddress) throws NamingCollisionException {
		if (deviceAddress == null) {
			throw new IllegalArgumentException("The device address may not be null");
		}
		String ifAndDevAddress;
		if (interfaceAddress == null || interfaceAddress.isEmpty()) {
			ifAndDevAddress = deviceAddress;
		}
		else {
			ifAndDevAddress = interfaceAddress + "#" + deviceAddress;
			if (deviceTasksByInterfaceAddress.get(deviceAddress) == null) {
				deviceTasksByInterfaceAddress.put(deviceAddress, new ArrayList<DeviceTask>());
			}
		}

		if (deviceConfigsByAddress.containsKey(ifAndDevAddress)) {
			throw new NamingCollisionException("Collision with interface/device addresses: " + ifAndDevAddress
					+ " in driver: " + id);
		}

		DeviceConfigImpl newDevice = new DeviceConfigImpl(interfaceAddress, deviceAddress, ifAndDevAddress, this);
		deviceConfigsByAddress.put(ifAndDevAddress, newDevice);

		return newDevice;
	}

	@Override
	public DeviceConfig getOrAddDevice(String interfaceAddress, String deviceAddress) {
		try {
			return addDevice(interfaceAddress, deviceAddress);
		} catch (NamingCollisionException e) {
			if (interfaceAddress == null) {
				return deviceConfigsByAddress.get(deviceAddress);
			}
			else {
				return deviceConfigsByAddress.get(interfaceAddress + "#" + deviceAddress);
			}
		}
	}

	@Override
	public DeviceConfig getDevice(String interfaceAddress, String deviceAddress) {
		if (interfaceAddress == null || interfaceAddress.isEmpty()) {
			return deviceConfigsByAddress.get(deviceAddress);
		}
		else {
			return deviceConfigsByAddress.get(interfaceAddress + "#" + deviceAddress);
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public Collection<DeviceConfig> getDevices() {
		return (Collection<DeviceConfig>) (Collection<?>) Collections.unmodifiableCollection(deviceConfigsByAddress
				.values());
	}

	@Override
	public void delete() {

		for (DeviceConfigImpl deviceConfig : deviceConfigsByAddress.values()) {
			deviceConfig.clear();
		}
		deviceConfigsByAddress.clear();
		rootConfigParent.removeDriver(this);
		rootConfigParent = null;
	}

	static void addDriverFromDomNode(Node driverConfigNode, RootConfig parentConfig) throws ParseException {

		String id = ChannelConfigImpl.getAttributeValue(driverConfigNode, "id");
		if (id == null) {
			throw new ParseException("driver has no id attribute");
		}

		DriverConfigImpl config;
		try {
			config = (DriverConfigImpl) parentConfig.addDriver(id);
		} catch (NamingCollisionException e) {
			throw new ParseException(e);
		}

		NodeList driverChildren = driverConfigNode.getChildNodes();

		for (int j = 0; j < driverChildren.getLength(); j++) {
			Node childNode = driverChildren.item(j);
			String childName = childNode.getNodeName();
			if (childName.equals("#text")) {
				continue;
			}
			else if (childName.equals("device")) {
				DeviceConfigImpl.addDeviceFromDomNode(childNode, config);
			}
			else if (childName.equals("readTimeout")) {
				config.readTimeout = ChannelConfigImpl.timeStringToMillis(childNode.getTextContent());
			}
			else if (childName.equals("connectTimeout")) {
				config.connectTimeout = ChannelConfigImpl.timeStringToMillis(childNode.getTextContent());
			}
			else if (childName.equals("connectRetryInterval")) {
				config.connectRetryInterval = ChannelConfigImpl.timeStringToMillis(childNode.getTextContent());
			}
			else if (childName.equals("disabled")) {
				String disabledString = childNode.getTextContent().toLowerCase();
				if (disabledString.equals("true")) {
					config.disabled = true;
				}
				else if (disabledString.equals("false")) {
					config.disabled = false;
				}
			}
			else {
				throw new ParseException("found unknown tag:" + childName);
			}

		}
	}

	Element getDomElement(Document document) {
		Element parentElement = document.createElement("driver");
		parentElement.setAttribute("id", id);

		Element childElement;

		if (readTimeout != null) {
			childElement = document.createElement("readTimeout");
			childElement.setTextContent(ChannelConfigImpl.millisToTimeString(readTimeout));
			parentElement.appendChild(childElement);
		}

		if (connectTimeout != null) {
			childElement = document.createElement("connectTimeout");
			childElement.setTextContent(ChannelConfigImpl.millisToTimeString(connectTimeout));
			parentElement.appendChild(childElement);
		}

		if (connectRetryInterval != null) {
			childElement = document.createElement("connectRetryInterval");
			childElement.setTextContent(ChannelConfigImpl.millisToTimeString(connectRetryInterval));
			parentElement.appendChild(childElement);
		}

		childElement = document.createElement("disabled");
		if (disabled) {
			childElement.setTextContent("true");
		}
		else {
			childElement.setTextContent("false");
		}
		parentElement.appendChild(childElement);

		for (DeviceConfigImpl deviceConfig : deviceConfigsByAddress.values()) {
			parentElement.appendChild(deviceConfig.getDomElement(document));
		}

		return parentElement;
	}

	void changeIfAndDevAddress(String oldIfAndDevAddress, String newIfAndDevAddress) throws NamingCollisionException {
		if (deviceConfigsByAddress.containsKey(newIfAndDevAddress)) {
			throw new NamingCollisionException("Collision with interface and device address:" + newIfAndDevAddress
					+ " in driver: " + id);
		}
		deviceConfigsByAddress.put(newIfAndDevAddress, deviceConfigsByAddress.remove(oldIfAndDevAddress));
	}

	void removeDevice(String address) {
		deviceConfigsByAddress.remove(address);
	}

	DriverConfigImpl clone(RootConfigImpl clonedParentConfig) {
		DriverConfigImpl configClone = new DriverConfigImpl(id, clonedParentConfig);
		configClone.readTimeout = readTimeout;
		configClone.connectTimeout = connectTimeout;
		configClone.connectRetryInterval = connectRetryInterval;
		configClone.disabled = disabled;
		for (Entry<String, DeviceConfigImpl> deviceConfig : deviceConfigsByAddress.entrySet()) {
			configClone.deviceConfigsByAddress.put(deviceConfig.getKey(), deviceConfig.getValue().clone(configClone));
		}
		return configClone;
	}

	DriverConfigImpl cloneWithDefaults(RootConfigImpl clonedParentConfig) {
		DriverConfigImpl configClone = new DriverConfigImpl(id, clonedParentConfig);

		if (readTimeout == null) {
			configClone.readTimeout = DriverConfig.READ_TIMEOUT_DEFAULT;
		}
		else {
			configClone.readTimeout = readTimeout;
		}

		if (connectTimeout == null) {
			configClone.connectTimeout = DriverConfig.CONNECT_TIMEOUT_DEFAULT;
		}
		else {
			configClone.connectTimeout = connectTimeout;
		}

		if (connectRetryInterval == null) {
			configClone.connectRetryInterval = DriverConfig.CONNECT_RETRY_INTERVAL_DEFAULT;
		}
		else {
			configClone.connectRetryInterval = connectRetryInterval;
		}

		configClone.disabled = disabled;

		for (Entry<String, DeviceConfigImpl> deviceConfig : deviceConfigsByAddress.entrySet()) {
			configClone.deviceConfigsByAddress.put(deviceConfig.getKey(),
					deviceConfig.getValue().cloneWithDefaults(configClone));
		}
		return configClone;
	}

}
