/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.openmuc.framework.driver.spi;

import java.io.IOException;
import java.util.List;

import org.openmuc.framework.config.ArgumentSyntaxException;
import org.openmuc.framework.config.ChannelScanInformation;
import org.openmuc.framework.config.DeviceScanInformation;

/**
 * 
 * The <code>DriverService</code> is the interface that all OpenMUC communication drivers have to implement and register
 * as a service in the OSGi environment. The OpenMUC Core Data Manager tracks this service and will therefore be
 * automatically notified of any new drivers in the framework. If sampling, listening or logging has been configured for
 * this driver, then the data manager will start right away with the appropriate actions.
 * 
 * The OpenMUC framework can give certain guarantees about the order of the functions it calls:
 * <ul>
 * <li>Communication related functions (e.g. connect,read,write..) are never called concurrently for the same device.</li>
 * <li>Communication related functions for different devices may be called concurrently by the framework unless an
 * interfaceAddress was configured. Thus drivers for communication protocols that do not support the parallel
 * communication to different communication devices over the same medium should make the configuration of an
 * interfaceAddress mandatory.</li>
 * <li>The framework calls read,listen,write or channelScan only if a the device is considered connected. The device is
 * only considered connected if the connect function has been called successfully.</li>
 * <li>Before a driver service is deregistered the framework calls disconnect for all connected devices. The disconnect
 * function should do any necessary resource clean up.</li>
 * </ul>
 * 
 * Some guidelines should be followed when implementing a driver for OpenMUC:
 * <ul>
 * <li>Logging may only be done with level <code>debug</code>. Even these debug messages should be used very sparsely.
 * Only use them to further explain errors that occur.</li>
 * <li>If the connection to a device is interrupted throw a <code>ConnectionException</code>. The framework will then
 * try to reconnect by calling the <code>connect</code> function.</li>
 * <li>All unchecked exceptions thrown by the <code>DriverService</code> are caught by the OpenMUC framework and logged
 * with level <code>error</code>. It is bad practice to throw these unchecked exceptions because they can clutter the
 * log file and slow down performance if the function is called many times. Instead the appropriate Flag should be
 * returned for the affected channels.</li>
 * </ul>
 * 
 */
public interface DriverService {

	/**
	 * Returns the ID of the driver. By convention the ID should be meaningful and all lower case letters (e.g. "mbus",
	 * "modbus").
	 * 
	 * @return the unique ID of the driver.
	 */
	public String getId();

	/**
	 * Returns a string with a description of what the driver does.
	 * 
	 * @return a string with a description of what the driver does
	 */
	public String getDescription();

	/**
	 * Scans for available communication devices and returns information about them. This operation is optional and may
	 * not be supported by specific drivers.
	 * 
	 * @param interfaceAddress
	 *            The interfaceAddress to be used for scanning.
	 * @param scanFilter
	 *            a string whose syntax depends on the specific communication driver being used. It tells the driver
	 *            where to scan for communication devices.
	 * @return A list of communication devices that were found. Returns an empty list if no device was found.
	 * @throws MethodNotSupportedException
	 *             if the method is not implemented by the driver
	 * @throws ArgumentSyntaxException
	 *             if an the scanFilter string cannot be understood by the driver
	 * @throws IOException
	 *             if an error occurs while scanning
	 */
	public List<DeviceScanInformation> scanForDevices(String interfaceAddress, String scanFilter, String settings,
			int timeout) throws UnsupportedOperationException, ArgumentSyntaxException, IOException;

	/**
	 * Scan a given communication device for available data channels. This operation is optional and my not be supported
	 * by specific drivers.
	 * 
	 * @param connection
	 *            ihe DeviceConnection of the communication device that is to be scanned for data channels.
	 * @return A list of data channels that were found.
	 * @throws MethodNotSupportedException
	 *             if the method is not implemented by the driver.
	 * @throws IOException
	 *             if an error occurs while scanning.
	 */
	public List<ChannelScanInformation> scanForChannels(DeviceConnection connection, int timeout)
			throws UnsupportedOperationException, ConnectionException;

	/**
	 * Attempts to connect to the given communication device using the given settings. The resulting connection resource
	 * object can be return by this function. The framework will then pass the connection object in all subsequent
	 * function calls meant for this device. If the syntax of the given interfaceAddress, deviceAddresse, or settings
	 * String is incorrect it will throw an <code>ArgumentSyntaxException</code>. If the connection attempt fails it
	 * throws a <code>ConnectionException</code>.
	 * 
	 * @param interfaceAddress
	 *            the configured interface address. Will be <code>null</code> if not configured.
	 * @param deviceAddress
	 *            the configured device address.
	 * @param settings
	 *            the settings that should be used for the communication with this device.
	 * @param timeout
	 *            the time in milliseconds after which the connection attempt shall be aborted and a
	 *            <code>ConnectionException</code> shall be thrown.
	 * @return the connection handle that is passed to subsequent read/listen/write/scanForChannels/disconnect function
	 *         calls.
	 * @throws ArgumentSyntaxException
	 *             if the syntax of interfaceAddress, deviceAddress or settings is incorrect.
	 * @throws ConnectionException
	 *             if the connection attempt fails.
	 */
	public Object connect(String interfaceAddress, String deviceAddress, String settings, int timeout)
			throws ArgumentSyntaxException, ConnectionException;

	/**
	 * Disconnects or closes the given connection. Cleans up any resources associated with the connection.
	 * 
	 * @param connection
	 *            the device connection that is to closed.
	 */
	public void disconnect(DeviceConnection connection);

	/**
	 * Reads the data channels that correspond to the given record containers. The read result is returned by setting
	 * the record in the containers. If for some reason no value can be read the record should be set anyways. In this
	 * case the record shall have a value and timestamp of <code>null</code> and a flag that best describes the reason
	 * of failure. If the connection to the device is interrupted, then any necessary resources that correspond to this
	 * connection should be cleaned up and a <code>ConnectionException</code> shall be thrown.
	 * 
	 * @param connection
	 *            the device connection that is to be used to read the channels.
	 * @param containers
	 *            the containers hold the information of what channels are to be read. They will be filled by this
	 *            function with the records read.
	 * @param containerListHandle
	 *            the containerListHandle returned by the last read call for this exact list of containers. Will be
	 *            equal to <cod>null</code> if this is the first read call for this container list after a connection
	 *            has been established. Driver implementations can optionally use this object to improve the read
	 *            performance.
	 * @param samplingGroup
	 *            the samplingGroup that was configured for this list of channels that are to be read. Sometimes it may
	 *            be desirable to give the driver a hint on how to group several channels when reading them. This can
	 *            done through the samplingGroup.
	 * @param timeout
	 *            the time after which the read function should interrupt the read process and return records with the
	 *            flag set to Flag.TIMEOUT. Note that OpenMUC Data Manager automatically sets the Flag of the latest
	 *            sample to Flag.TIMEOUT once the time has elapsed without the read function returning. Thus even if
	 *            driver neglects the timeout the latest sample will always be flagged correctly. The main disadvantage
	 *            of ignoring the timeout is that the driver could potentially block new read, write or disconnect
	 *            requests from executing.
	 * @return the containerListHandle Object that will passed the next time the same list of channels is to be read.
	 *         Use this Object as a handle to improve performance or simply return <code>null</code>.
	 * @throws UnsupportedOperationException
	 *             if the method is not implemented by the driver.
	 * @throws ConnectionException
	 *             if the connection to the device was interrupted.
	 */
	public Object read(DeviceConnection connection, List<ChannelRecordContainer> containers,
			Object containerListHandle, String samplingGroup, int timeout) throws UnsupportedOperationException,
			ConnectionException;

	/**
	 * Starts listening on the given connection for data from the channels that correspond to the given record
	 * containers. Will overwrite all data channels set by the previous startListening call. Will notify the given
	 * listener of new records that arrive on the data channels.
	 * 
	 * @param connection
	 *            the device connection that is to be used to listen.
	 * @param containers
	 *            the containers identify the channels to listen on. They will be filled by this function with the
	 *            records received and passed to the listener.
	 * @param listener
	 *            the listener to inform that new data has arrived.
	 * @throws UnsupportedOperationException
	 *             if the method is not implemented by the driver.
	 * @throws ConnectionException
	 *             if the connection to the device was interrupted.
	 */
	public void startListening(DeviceConnection connection, List<ChannelRecordContainer> containers,
			RecordsReceivedListener listener) throws UnsupportedOperationException, ConnectionException;

	/**
	 * Writes the data channels that correspond to the given value containers. The write result is returned by setting
	 * the flag in the containers. If the connection to the device is interrupted, then any necessary resources that
	 * correspond to this connection should be cleaned up and a <code>ConnectionException</code> shall be thrown.
	 * 
	 * @param connection
	 *            the device connection that is to be used to write the channels.
	 * @param containers
	 *            the containers hold the information of what channels are to be written and the values that are to
	 *            written. They will be filled by this function with a flag stating whether the write process was
	 *            successful or not.
	 * @param containerListHandle
	 *            the containerListHandle returned by the last write call for this exact list of containers. Will be
	 *            equal to <cod>null</code> if this is the first read call for this container list after a connection
	 *            has been established. Driver implementations can optionally use this object to improve the write
	 *            performance.
	 * @param timeout
	 *            see read for detailed documentation.
	 * @return the containerListHandle Object that will passed the next time the same list of channels is to be written.
	 *         Use this Object as a handle to improve performance or simply return <code>null</code>.
	 * @throws UnsupportedOperationException
	 *             if the method is not implemented by the driver.
	 * @throws ConnectionException
	 *             if the connection to the device was interrupted.
	 */
	public Object write(DeviceConnection connection, List<ChannelValueContainer> containers,
			Object containerListHandle, int timeout) throws UnsupportedOperationException, ConnectionException;

}
