/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.framework.datalogger.ascii;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Calendar;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import org.openmuc.framework.data.ByteStringValue;
import org.openmuc.framework.data.DoubleValue;
import org.openmuc.framework.data.Record;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class FileInHandler {

	private static final Logger logger = LoggerFactory.getLogger(FileOutHandler.class);

	private final String path;
	private final String extension;

	public FileInHandler(String path, String extension) {
		this.path = path;
		this.extension = extension;
	}

	public int findIndexInHeader(String id, String line) {
		String tokens[] = line.split(" ");
		for (int i = 2; i < tokens.length; i++) {
			if (id.equals(tokens[i])) {
				return i;
			}
		}
		return -1;
	}

	public List<Record> readRecords(String id, long startTimestamp, long endTimestamp) {
		File folder = new File(path);
		if (!folder.exists() || !folder.isDirectory()) {
			logger.error("Cannot read history because " + path + " is no directory or does not exist");
			return Collections.emptyList();
		}

		Calendar date = Calendar.getInstance();
		date.setTimeInMillis(startTimestamp);

		Calendar until = Calendar.getInstance();
		until.setTimeInMillis(endTimestamp);

		List<String> filenames = new LinkedList<String>();
		while (date.before(until)) {
			filenames.add(String.format("%02d%02d%02d%s", date.get(Calendar.YEAR) % 100, date.get(Calendar.MONTH) + 1,
					date.get(Calendar.DAY_OF_MONTH), extension));
			date.add(Calendar.DAY_OF_MONTH, 1);
		}

		List<Record> result = new LinkedList<Record>();
		try {
			while (filenames.isEmpty() == false) {
				File inFile = new File(path, filenames.remove(0));
				if (inFile.exists()) {
					readFromFile(inFile, id, result, startTimestamp, endTimestamp);
				}
			}
		} catch (FileNotFoundException e) {
			logger.debug("Error on reading file. File not found");
		} catch (IOException e) {
			logger.debug("IOException on reading file...", e);
		}

		return result;
	}

	private void readFromFile(File inFile, String id, List<Record> result, long startTimestamp, long endTimestamp)
			throws NumberFormatException, IOException {
		int index = -1;
		BufferedReader in = new BufferedReader(new FileReader(inFile));

		while (in.ready()) {
			String line = in.readLine();
			if (line.startsWith("##")) {
				index = findIndexInHeader(id, line);
			}
			else if (line.startsWith("#")) {
				// ignore comment lines
			}
			else if (index != -1) {
				String tokens[] = line.split(" ");
				long timestamp = Long.parseLong(tokens[1]) * 1000;
				if (startTimestamp <= timestamp && timestamp <= endTimestamp) {
					if (tokens[index].equals("-99.9")) {
						// ignore line
					}
					else if (tokens[index].startsWith("[") && tokens[index].endsWith("]")) {
						byte[] value = readByteArray(tokens[index]);
						result.add(new Record(new ByteStringValue(value), timestamp));
					}
					else {
						double val = Double.parseDouble(tokens[index]);
						result.add(new Record(new DoubleValue(val), timestamp));
					}
				}
			}
		}
	}

	private byte[] readByteArray(String array) {
		array = array.replaceAll("([|])", "");
		byte result[] = new byte[array.length() / 2];
		for (int i = 0; i < result.length; i++) {
			result[i] = (byte) Integer.parseInt(array.substring(i * 2, (i + 1) * 2), 16);
		}
		return result;
	}
}
