/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.openmuc.framework.driver.modbustcp;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Hashtable;
import java.util.List;

import net.wimpi.modbus.ModbusException;

import org.openmuc.framework.config.ArgumentSyntaxException;
import org.openmuc.framework.config.ChannelScanInformation;
import org.openmuc.framework.config.DeviceScanInformation;
import org.openmuc.framework.data.Flag;
import org.openmuc.framework.data.Record;
import org.openmuc.framework.data.Value;
import org.openmuc.framework.driver.modbustcp.ModbusTcpChannel.EAccess;
import org.openmuc.framework.driver.spi.ChannelRecordContainer;
import org.openmuc.framework.driver.spi.ChannelValueContainer;
import org.openmuc.framework.driver.spi.ConnectionException;
import org.openmuc.framework.driver.spi.DeviceConnection;
import org.openmuc.framework.driver.spi.DriverService;
import org.openmuc.framework.driver.spi.RecordsReceivedListener;
import org.osgi.service.component.ComponentContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public final class ModbusTcpDriverImpl implements DriverService {

	private final static Logger logger = LoggerFactory.getLogger(ModbusTcpDriverImpl.class);

	/**
	 * List do manage ModbusTcpChannel Objects to avoid to check the syntax of each channel address for every read or
	 * write
	 */
	private Hashtable<String, ModbusTcpChannel> modbusTcpChannels;

	protected void activate(ComponentContext context) {
		modbusTcpChannels = new Hashtable<String, ModbusTcpChannel>();
		logger.debug("Activated Modbus Driver");
	}

	protected void deactivate(ComponentContext context) {
		// TODO Disconnect of all devices
		// TODO maintain an internal device list
		logger.debug("Deactivated Modbus Driver");
	}

	@Override
	public String getId() {
		return "modbustcp";
	}

	@Override
	public String getDescription() {
		return "Driver for ModbusTCP";
	}

	@Override
	public Object connect(String interfaceAddress, String deviceAddress, String settings, int timeout)
			throws ConnectionException {

		ModbusTcpDevice device = new ModbusTcpDevice(deviceAddress);
		device.connect();

		return device;
	}

	@Override
	public void disconnect(DeviceConnection connection) {
		((ModbusTcpDevice) connection.getConnectionHandle()).disconnect();
	}

	@Override
	public Object read(DeviceConnection connection, List<ChannelRecordContainer> containers,
			Object containerListHandle, String samplingGroup, int timeout) throws UnsupportedOperationException {

		// TODO add Throw ConnectionException --> ModbusIOException?

		ModbusTcpDevice device = (ModbusTcpDevice) connection.getConnectionHandle();

		if (samplingGroup.isEmpty()) {
			// reads channels one by one
			return readChannels(device, containers, timeout);
		}
		else {
			// reads whole samplingGroup at once
			return readChannelGroup(device, containers, containerListHandle, samplingGroup, timeout);
		}

	}

	@Override
	public Object write(DeviceConnection connection, List<ChannelValueContainer> containers,
			Object containerListHandle, int timeout) throws UnsupportedOperationException, ConnectionException {

		ModbusTcpDevice device = (ModbusTcpDevice) connection.getConnectionHandle();

		for (ChannelValueContainer container : containers) {

			ModbusTcpChannel modbusChannel = getModbusTcpChannel(container.getChannelAddress(), EAccess.WRITE);
			if (modbusChannel != null) {
				try {
					device.writeChannel(modbusChannel, container.getValue());
				} catch (ModbusException modbusException) {
					modbusException.printStackTrace();
					throw new ConnectionException("Unable to write data on channel address: "
							+ container.getChannelAddress());
				} catch (Exception e) {
					e.printStackTrace();
					logger.error("Unable to write data on channel address: " + container.getChannelAddress());
				}
			}
			else {
				// TODO
				logger.error("Unable to write data on channel address: " + container.getChannelAddress()
						+ "ModbusTcpChannel = null");
			}
		}

		return null;

	}

	/**
	 * Reads all requested channels one by one.
	 * 
	 * @param device
	 * @param containers
	 * @param timeout
	 * @return
	 */
	private Object readChannels(ModbusTcpDevice device, List<ChannelRecordContainer> containers, int timeout) {

		new Date().getTime();

		for (ChannelRecordContainer container : containers) {
			long receiveTime = System.currentTimeMillis();
			ModbusTcpChannel channel = getModbusTcpChannel(container.getChannelAddress(), EAccess.READ);
			Value value;
			try {
				value = device.readChannel(channel, timeout);

				// logger.debug(channel.getChannelAddress() + ": value = " + value.asDouble());
				logger.debug(channel.getChannelAddress() + ": value = '" + value.toString() + "'");

				container.setRecord(new Record(value, receiveTime));
			} catch (ModbusException e) {
				e.printStackTrace();
				container.setRecord(new Record(null, null, Flag.DRIVER_ERROR_CHANNEL_NOT_ACCESSIBLE));
			}
		}

		// logger.debug("### readChannels duration in ms = " + ((new Date().getTime()) - startTime));

		return null;
	}

	/**
	 * Read the all channels of the samplingGroup at once.
	 * 
	 * @param device
	 * @param containers
	 * @param containerListHandle
	 * @param samplingGroup
	 * @param timeout
	 * @return
	 */
	private Object readChannelGroup(ModbusTcpDevice device, List<ChannelRecordContainer> containers,
			Object containerListHandle, String samplingGroup, int timeout) {

		new Date().getTime();

		// NOTE: containerListHandle is null if something changed in configuration!!!

		ModbusTcpChannelGroup channelGroup = null;

		// use existing channelGroup
		if (containerListHandle != null) {
			if (containerListHandle instanceof ModbusTcpChannelGroup) {
				channelGroup = (ModbusTcpChannelGroup) containerListHandle;
			}
		}

		// create new channelGroup
		if (channelGroup == null) {
			ArrayList<ModbusTcpChannel> channelList = new ArrayList<ModbusTcpChannel>();
			for (ChannelRecordContainer container : containers) {
				channelList.add(getModbusTcpChannel(container.getChannelAddress(), EAccess.READ));
			}
			channelGroup = new ModbusTcpChannelGroup(samplingGroup, channelList);
		}

		// read all channels of the group
		try {
			device.readChannelGroup(channelGroup, containers, timeout);

		} catch (ModbusException e) {
			e.printStackTrace();

			// set channel values and flag, otherwise the datamanager will throw a null pointer exception
			// and the framework collapses.
			device.setChannelsWithErrorFlag(containers);
		}

		// logger.debug("### readChannelGroup duration in ms = " + ((new Date().getTime()) - startTime));

		return channelGroup;
	}

	private ModbusTcpChannel getModbusTcpChannel(String channelAddress, EAccess access) {

		ModbusTcpChannel modbusTcpChannel = null;

		// check if the channel object already exists in the list
		if (modbusTcpChannels.containsKey(channelAddress)) {
			modbusTcpChannel = modbusTcpChannels.get(channelAddress);

			// if the channel object exists the access flag might has to be updated
			// (this is case occurs when the channel is readable and writable)
			if (!modbusTcpChannel.getAccessFlag().equals(access)) {
				modbusTcpChannel.update(access);
			}
		}
		// create a new channel object
		else {
			modbusTcpChannel = new ModbusTcpChannel(channelAddress, access);
			modbusTcpChannels.put(channelAddress, modbusTcpChannel);
		}

		return modbusTcpChannel;

	}

	@Override
	public void startListening(DeviceConnection connection, List<ChannelRecordContainer> containers,
			RecordsReceivedListener listener) throws UnsupportedOperationException {
		// not supported
		throw new UnsupportedOperationException();
	}

	@Override
	public List<DeviceScanInformation> scanForDevices(String interfaceAddress, String scanFilter, String settings,
			int timeout) throws UnsupportedOperationException, ArgumentSyntaxException, IOException {
		// not supported
		throw new UnsupportedOperationException();
	}

	@Override
	public List<ChannelScanInformation> scanForChannels(DeviceConnection connection, int timeout)
			throws UnsupportedOperationException, ConnectionException {
		// not supported
		throw new UnsupportedOperationException();
	}
}
