/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.framework.server.restws.channel;

import java.io.IOException;

import org.openmuc.framework.data.Record;
import org.openmuc.framework.data.Value;
import org.openmuc.framework.dataaccess.Channel;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.representation.EmptyRepresentation;
import org.restlet.representation.Representation;
import org.restlet.representation.StringRepresentation;
import org.restlet.resource.Get;
import org.restlet.resource.Put;
import org.restlet.resource.ServerResource;

import com.fasterxml.jackson.core.JsonProcessingException;

/**
 * Webservice for accessing single channels
 * 
 * @author Karsten Mueller-Bier
 */
public final class ChannelResource extends ServerResource {

	/**
	 * Get the last record of a single channel formatted in JSON
	 * 
	 * @return A StringRepresentation containing the JSON string
	 * @throws JsonProcessingException
	 *             If there was a problem with creating the JSON string
	 */
	@Get("json")
	public Representation getJson() throws IOException {
		Channel channel = WebserviceProvider.getChannel(getRequest(), getResponse());

		if (channel == null) {
			return null;
		}
		Record lastSample = channel.getLatestSample();

		JsonWrapper wrapper = WebserviceProvider.getWrapper();
		return new StringRepresentation(wrapper.fromRecord(lastSample), MediaType.APPLICATION_JSON);
	}

	/**
	 * Set a new record for a single channel.
	 * 
	 * This method will only process the request if it came through a encyripted communication and the user
	 * authenticates itself using HTTP_BASIC authentication
	 * 
	 * @param entity
	 *            The transferred JSON String
	 * @return A StringRepresentation containing the JSON string of the new record
	 * @throws IOException
	 *             If there was a problem with reading the JSON string
	 */
	@Put("json")
	public Representation setFromJson(Representation entity) throws IOException {

		// Check if communication is encrypted (e.g. with HTTPS)
		if (!getRequest().isConfidential()) {
			getResponse().setStatus(Status.CLIENT_ERROR_FORBIDDEN);
			return new StringRepresentation("Use HTTPS to update values\n");
		}

		// Communication is encrypted, check if user authenticates itself
		if (!WebserviceProvider.authenticate(getRequest(), getResponse())) {
			return new EmptyRepresentation();
		}

		JsonWrapper wrapper = WebserviceProvider.getWrapper();
		Value value = wrapper.toValue(entity.getStream());

		Channel channel = WebserviceProvider.getChannel(getRequest(), getResponse());
		Record newSample = new Record(value, System.currentTimeMillis());
		channel.setLatestSample(newSample);

		return new StringRepresentation(wrapper.fromRecord(channel.getLatestSample()), MediaType.APPLICATION_JSON);
	}
}
