/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.framework.server.restws.channel;

import java.io.IOException;
import java.io.InputStream;
import java.util.List;

import org.openmuc.framework.data.ByteStringValue;
import org.openmuc.framework.data.DoubleValue;
import org.openmuc.framework.data.Record;
import org.openmuc.framework.data.TypeConversionException;
import org.openmuc.framework.data.Value;
import org.openmuc.framework.dataaccess.Channel;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.fasterxml.jackson.databind.node.ArrayNode;
import com.fasterxml.jackson.databind.node.ObjectNode;

/**
 * Helper class that transforms the objects communicated with the webservices in this package to and from JSON
 * 
 * @author Karsten Mueller-Bier
 */
public final class JsonWrapper {

	private final ObjectMapper mapper;

	public JsonWrapper() {
		mapper = new ObjectMapper();

		// INDENT_OUTPUT will arrange the outgoing json strings so that they are easier to read for humans
		// Removing or commenting out this line will print the whole json string in one line with as less whitespace as
		// possible, reducing bandwith and readability
		mapper.enable(SerializationFeature.INDENT_OUTPUT);
	}

	/**
	 * Creates a JSON string from a single Record object
	 * 
	 * @see JsonWrapper#writeRecordToNode(Record, ObjectNode)
	 * @param value
	 *            The record to transform
	 * @return JSON representation of the record
	 * @throws JsonProcessingException
	 *             Error on processing ObjectNode. Unlikely that it happens
	 */
	public String fromRecord(Record value) throws JsonProcessingException {
		ObjectNode root = mapper.createObjectNode();

		writeRecordToNode(value, root);

		return mapper.writeValueAsString(root);
	}

	/**
	 * Creates a JSON string containing all records given
	 * 
	 * @param values
	 *            The records to transform
	 * @return JSON representation of the records
	 * @throws JsonProcessingException
	 *             Error on processing ObjectNode. Still unlikely that it happens
	 */
	public String fromRecordList(List<Record> values) throws JsonProcessingException {
		ArrayNode root = mapper.createArrayNode();

		for (Record record : values) {
			ObjectNode node = root.addObject();
			writeRecordToNode(record, node);
		}

		return mapper.writeValueAsString(root);
	}

	/**
	 * Creates a JSON string containing all channels given
	 * 
	 * @see JsonWrapper#writeRecordToNode(Record, ObjectNode)
	 * @param channels
	 *            List of channels to write
	 * @return JSON representation of the channels
	 * @throws JsonProcessingException
	 *             Error on processing ObjectNode. Unlikely that this exception is thrown
	 */
	public String fromChannelList(List<Channel> channels) throws JsonProcessingException {
		ArrayNode root = mapper.createArrayNode();

		for (Channel channel : channels) {
			ObjectNode node = root.addObject();
			node.put("label", channel.getLabel());
			node = node.with("record");
			if (channel.getLatestSample() != null) {
				writeRecordToNode(channel.getLatestSample(), node);
			}
		}

		return mapper.writeValueAsString(root);
	}

	/**
	 * Converts the value field of this JSON string to a org.openmuc.framework.data.Value object
	 * 
	 * @param textStream
	 *            Stream containing the JSON string
	 * @return The converted Value object
	 * @throws JsonProcessingException
	 *             If there is a problem extracting the value out of the JSON string
	 * @throws IOException
	 *             If there is a problem extracting the JSON String out of the InputStream
	 */
	public Value toValue(InputStream textStream) throws JsonProcessingException, IOException {
		JsonNode root = mapper.readTree(textStream);

		Value result = null;
		if (root.has("value")) {
			JsonNode valueNode = root.get("value");

			if (valueNode.isNumber()) {
				result = new DoubleValue(valueNode.asDouble());
			}
			else {
				result = new ByteStringValue(valueNode.binaryValue());
			}
		}

		return result;
	}

	/**
	 * Writes a Record object into a JSON ObjectNode. The timestamp and flag are always written, the value is omitted of
	 * there is no value present
	 * 
	 * @param record
	 *            The Record object to transform
	 * @param node
	 *            The JSON ObjectNode that shall contain the record
	 */
	private void writeRecordToNode(Record record, ObjectNode node) {
		node.put("timestamp", record.getTimestamp());
		node.put("flag", record.getFlag().name());
		if (record.getValue() != null) {
			try {
				node.put("value", record.getValue().asDouble());
			} catch (TypeConversionException e) {
				node.put("value", record.getValue().asByteArray());
			}
		}
	}
}
