/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.framework.server.restws.channel;

import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.util.Dictionary;

import org.openmuc.framework.authentication.AuthenticationService;
import org.openmuc.framework.dataaccess.Channel;
import org.openmuc.framework.dataaccess.DataAccessService;
import org.osgi.service.http.HttpContext;
import org.restlet.Application;
import org.restlet.Context;
import org.restlet.Request;
import org.restlet.Response;
import org.restlet.data.ChallengeScheme;
import org.restlet.data.Status;
import org.restlet.ext.osgi.IApplicationProvider;
import org.restlet.routing.Router;
import org.restlet.security.ChallengeAuthenticator;
import org.restlet.security.SecretVerifier;
import org.restlet.security.Verifier;

/**
 * Initial Point of the RESTful channel webservice. On starting this bundle, its createApplication methode is
 * automatically called by org.restlet.ext.osgi.servlet.RestletServletService.
 * 
 * Calling createApplication results in a restlet Application containing the channel and channellist webservice.
 * 
 * This class also stores all helper classes and services needed by ChannelListService and ChannelService
 * 
 * @author Karsten Mueller-Bier
 */
public final class WebserviceProvider implements IApplicationProvider {

	private static volatile JsonWrapper wrapper = null;
	private static AuthenticationService authenticatorService = null;
	private static DataAccessService dataAccess = null;

	private static ChallengeAuthenticator auth;

	private volatile Application app = null;

	/**
	 * @return The root name of all services of this application
	 */
	@Override
	public String getAlias() {
		return "/rest";
	}

	@Override
	public Application createApplication(Context context) {
		if (app == null) {
			synchronized (this) {
				if (app == null) {
					app = new Application();
					app.setContext(context);

					Router router = new Router(context);

					router.attach("/channel/{channelLabel}/history", ChannelHistoryResource.class);
					router.attach("/channel/{channelLabel}", ChannelResource.class);
					router.attach("/channel/", ChannelListResource.class);
					router.attach("/channel", ChannelListResource.class);

					auth = createAuthenticator(context);
					auth.setNext(router);

					app.setInboundRoot(auth);
				}
			}
		}

		return app;
	}

	@Override
	public HttpContext getContext() {
		// use standard HTTP context
		return null;
	}

	@Override
	public Dictionary<String, Object> getInitParms() {
		// use standard init parameter
		return null;
	}

	protected void setDataAccessService(DataAccessService dataAccessService) {
		dataAccess = dataAccessService;
	}

	protected void unsetDataAccessService(DataAccessService dataAccessService) {
		dataAccess = null;
	}

	protected void setAuthenticationService(AuthenticationService authenticationService) {
		authenticatorService = authenticationService;
	}

	protected void unsetAuthenticationService(AuthenticationService authenticationService) {
		authenticatorService = null;
	}

	static DataAccessService getDataAccess() {
		if (dataAccess == null) {
			throw new NullPointerException();
		}

		return dataAccess;
	}

	static JsonWrapper getWrapper() {
		if (wrapper == null) {
			wrapper = new JsonWrapper();
		}

		return wrapper;
	}

	/**
	 * Helper method that reads the requested channel from the URL. If no channel with that label exists, the HTTP
	 * response is set to 400 (Bad Request)
	 * 
	 * @return The requested channel, null if no such channel exists
	 */
	static Channel getChannel(Request request, Response response) {
		String label = (String) request.getAttributes().get("channelLabel");
		try {
			label = URLDecoder.decode(label, "UTF-8");
		} catch (UnsupportedEncodingException e) {
			// Cannot happen. The character set UTF-8 is mandatory for all Java Runtime Environments
			// See: http://docs.oracle.com/javase/1.4.2/docs/api/java/nio/charset/Charset.html
		}

		Channel channel = getDataAccess().getChannel(label);

		if (channel == null) {
			response.setStatus(Status.CLIENT_ERROR_BAD_REQUEST);
			return null;
		}

		return channel;
	}

	/**
	 * Helper method to check if the incoming request is made with valid credentials
	 * 
	 * @param request
	 *            Request object of this client request
	 * @param response
	 *            Response to this client request
	 * @return true if client is authenticated
	 */
	static boolean authenticate(Request request, Response response) {
		if (!request.getClientInfo().isAuthenticated()) {
			auth.challenge(response, false);
			return false;
		}
		return true;
	}

	/**
	 * Creating the ChallengeAuthenticator guarding sensible methods. This authenticator uses the same Mechanism and
	 * user/password file as the WebUI-Base bundle. The user and password will be transferred using HTTP_BASIC, meaning
	 * that both user name and password will be sent in unencrypted. Encryption has to be implemented by the
	 * transporting protocol (e.g. HTTPS with SSL)
	 * 
	 * @param context
	 *            The HttpContext of the application
	 * @return The created ChallengeAuthenticator
	 */
	private ChallengeAuthenticator createAuthenticator(Context context) {
		boolean optional = true;
		ChallengeScheme challengeScheme = ChallengeScheme.HTTP_BASIC;
		String realm = "OpenMUC";

		Verifier verifier = new SecretVerifier() {

			@Override
			public int verify(String identifier, char[] secret) {
				String password = new String(secret);
				if (authenticatorService.login(identifier, password)) {
					return RESULT_VALID;
				}
				return RESULT_INVALID;
			}
		};

		return new ChallengeAuthenticator(context, optional, challengeScheme, realm, verifier);
	}
}
