package org.openmuc.android.fragments;

import java.lang.ref.WeakReference;
import java.util.Arrays;
import java.util.List;
import java.util.Set;

import org.openmuc.android.activities.handheld.DeleteChannelActivity;
import org.openmuc.android.rest.data.Channel;
import org.openmuc.android.util.ChannelAdapter;
import org.openmuc.android.util.MainMultiChoiceModeListener;
import org.openmuc.android.util.MainMultiChoiceModeListener.ChannelDeleteChoiceCallback;

import android.annotation.SuppressLint;
import android.app.Activity;
import android.content.Intent;
import android.os.Build;
import android.os.Bundle;
import android.os.Handler;
import android.os.Message;
import android.os.Parcelable;
import android.support.v4.app.ListFragment;
import android.view.View;
import android.widget.AbsListView;
import android.widget.AdapterView;
import android.widget.AdapterView.OnItemLongClickListener;
import android.widget.ListView;

/**
 * Fragment that displays a list of all registered channels provided from the activity by calling setChannelList. <br>
 * <br>
 * An activity that uses this fragment must implement the {@link ChannelListFragmentCallback} <br>
 * <br>
 * In devices running Android version 3.0 or higher, a MultiChoiceModeListener to delete channels is also added here.
 * 
 * @author Karsten Müller-Bier
 */
public class ChannelListFragment extends ListFragment implements ChannelDeleteChoiceCallback {

	private ChannelAdapter adapter;
	private List<Channel> channelsToRead;
	private final Handler handler = new AdapterHandler(this);

	private ChannelListFragmentCallback callback;

	/**
	 * @param list
	 *            The list of Channel object that are to be displayed by this fragment
	 */
	public void setChannelList(List<Channel> list) {
		channelsToRead = list;
	}

	/**
	 * @return A handler of this fragment, that informs this fragment to redraw itself
	 * @see AdapterHandler
	 */
	public Handler getAdapterHandler() {
		return handler;
	}

	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);

		if (savedInstanceState != null) {
			Parcelable[] data = savedInstanceState.getParcelableArray("channels");
			if (data != null && data.length != 0) {
				Channel channels[] = (Channel[]) data;
				channelsToRead = Arrays.asList(channels);
			}
		}
	}

	@Override
	public void onAttach(Activity activity) {
		super.onAttach(activity);

		if (activity instanceof ChannelListFragmentCallback) {
			callback = (ChannelListFragmentCallback) activity;
		}
		else {
			throw new IllegalArgumentException(activity.getClass().getName() + " does not implement "
					+ ChannelListFragmentCallback.class.getName());
		}
	}

	@Override
	public void onDetach() {
		super.onDetach();
		callback = null;
	}

	@SuppressLint("NewApi")
	@Override
	public void onStart() {
		super.onStart();

		adapter = new ChannelAdapter(getActivity(), channelsToRead);
		setListAdapter(adapter);
		setEmptyText("No channels selected");

		if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.HONEYCOMB) {
			getListView().setChoiceMode(AbsListView.CHOICE_MODE_MULTIPLE_MODAL);
			getListView().setMultiChoiceModeListener(new MainMultiChoiceModeListener(this, getListView(), adapter));
		}
		else {
			getListView().setOnItemLongClickListener(new OnItemLongClickListener() {
				@Override
				public boolean onItemLongClick(AdapterView<?> paramAdapterView, View paramView, int paramInt,
						long paramLong) {
					Intent deleteIntent = new Intent(getActivity(), DeleteChannelActivity.class);
					getActivity().startActivityForResult(deleteIntent, DeleteChannelActivity.REQUEST_CODE);
					return true;
				}
			});
		}
	}

	@Override
	public void onListItemClick(ListView l, View v, int position, long id) {
		Channel c = (Channel) l.getItemAtPosition(position);
		callback.channelSelected(c);
	}

	@Override
	public void onSaveInstanceState(Bundle outState) {
		super.onSaveInstanceState(outState);

		outState.putParcelableArray("channels", channelsToRead.toArray(new Channel[0]));
	}

	/**
	 * Handler workaround to let other objects notify the {@link ChannelListFragment} when the data set has been updated <br>
	 * <br>
	 * This workaround is necessary, as references of Fragment classes must only be known to the activity that created
	 * it. Leaking the reference to another object breaks the lifecycle of the fragment. To work around that, the
	 * Handler only holds a weak reference of the fragment, that doesn't disturb its lifecycle
	 * 
	 * @author Karsten Müller-Bier
	 */
	public static class AdapterHandler extends Handler {
		private final WeakReference<ChannelListFragment> outer;

		public AdapterHandler(ChannelListFragment outer) {
			this.outer = new WeakReference<ChannelListFragment>(outer);
		}

		@Override
		public void handleMessage(Message msg) {
			if (outer != null && outer.get() != null) {
				outer.get().adapter.notifyDataSetChanged();
			}
			else {
				super.handleMessage(msg);
			}
		}
	}

	/**
	 * Interface used by {@link ChannelListFragment} to notify its activity about user interaction
	 * 
	 * @author Karsten Müller-Bier
	 */
	public interface ChannelListFragmentCallback {
		/**
		 * Callback notifying the activity that the user selected a channel
		 * 
		 * @param selectedChannel
		 *            the selected channel
		 */
		void channelSelected(Channel selectedChannel);

		/**
		 * Calling the activity that certain channels shall be removed from the list
		 * 
		 * @param channelsToDelete
		 *            the channels to remove
		 */
		void deleteChannel(Set<String> channelsToDelete);
	}

	@Override
	public void deleteChannels(Set<String> channelsToDelete) {
		callback.deleteChannel(channelsToDelete);
	}
}
